from ordf.graph import Graph
from ordf.vocab.owl import predicate, object_predicate, \
    Class, Individual, AnnotatibleTerms
from logging import getLogger

log = getLogger(__name__)
skos_rdf = "http://www.w3.org/TR/skos-reference/skos.rdf"

from ordf.namespace import SKOS

class _Common(object):
    prefLabel = predicate(SKOS.prefLabel)
    altLabel = predicate(SKOS.altLabel)

class ConceptScheme(Class, _Common):
    topConcept = object_predicate(SKOS.topConcept, "Concept", globals())

    def __init__(self, *av, **kw):
        kwa = kw.copy()
        kwa["skipClassMembership"] = True
        kwa.setdefault("factoryClass", Concept)
        super(ConceptScheme, self).__init__(*av, **kwa)
        if not kw.get("skipClassMembership"):
            self.type = SKOS.ConceptScheme

class Concept(Class, _Common):
    inScheme = object_predicate(SKOS.inScheme, ConceptScheme)
    topConceptOf = object_predicate(SKOS.topConceptOf, ConceptScheme)
    notation = predicate(SKOS.notation)
    narrower = object_predicate(SKOS.narrower, "Concept", globals())
    broader = object_predicate(SKOS.broader, "Concept", globals())

    def __init__(self, *av, **kw):
        kwa = kw.copy()
        kwa["skipClassMembership"] = True
        kwa.setdefault("factoryClass", Concept)
        super(Concept, self).__init__(*av, **kwa)
        if not kw.get("skipClassMembership"):
            self.type = SKOS.Concept

class NotationCache(object):
    def __init__(self, store):
        self.cache = {}
        self.store = store
    def get(self, resource):
        from telescope import Select, v
        notation = self.cache.get(resource)
        if notation is None:
            q = Select([v.notation], distinct=True).where(
                (resource, SKOS.notation, v.notation)
                )
            for notation in self.store.query(q.compile()):
                if isinstance(notation, list): notation = notation[0]
                self.cache[resource] = notation
                break
        return notation

def rdf_data():
    log.info("Fetching %s" % skos_rdf)
    g = Graph(identifier=skos_rdf).parse(skos_rdf)
    log.info("Parsed %s (%d triples)" % (skos_rdf, len(g)))
    yield g

def inference_rules(handler, network):
    from FuXi.DLP.DLNormalization import NormalFormReduction
    skos = handler.get(skos_rdf)
    if len(skos) == 0:
        for skos in rdf_data():
            handler.put(skos)
    NormalFormReduction(skos)
    return network.setupDescriptionLogicProgramming(skos, addPDSemantics=False)



from ordf.graph import Graph
from logging import getLogger

log = getLogger(__name__)
rss_rdf = "http://purl.org/rss/1.0/schema.rdf"

def rdf_data():
    log.info("Fetching %s" % rss_rdf)
    g = Graph(identifier=rss_rdf).parse(rss_rdf)
    log.info("Parsed %s (%d triples)" % (rss_rdf, len(g)))
    yield g

def inference_rules(handler, network):
    from FuXi.DLP.DLNormalization import NormalFormReduction
    rss = handler.get(rss_rdf)
    if len(rss) == 0:
        for rss in rdf_data():
            handler.put(rss)
    NormalFormReduction(rss)
    return network.setupDescriptionLogicProgramming(rss, addPDSemantics=False)



"""
.. autoclass:: FuXiReasoner
   :show-inheritance:
"""

from ordf.graph import Graph, _Graph, ConjunctiveGraph
from ordf.handler import Handler, HandlerPlugin
from ordf.namespace import CS, DC, FOAF, ORDF, RDF
from ordf.term import URIRef
from ordf.vocab.foaf import Agent

from FuXi.Rete.Network import ReteNetwork
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.Horn.HornRules import HornFromN3
from FuXi.Rete.Util import generateTokenSet
from FuXi.Syntax.InfixOWL import Class, Property
from StringIO import StringIO
from getpass import getuser
from string import strip

from logging import getLogger
log = getLogger(__name__)

class FuXiReasoner(HandlerPlugin):
    """
    This handler implements a production rule reasoning system. In
    normal operation it would listen to a one queue using
    :class:`~ordf.handler.queue.RabbitHandler` and the handler would
    be configured to write to a different exchange using
    :class:`~ordf.handler.queue.RabbitQueue`.

    When the :meth:put method is called, it will be with a store that
    already contains a :class:`~ordf.changeset.ChangeSet`. This handler
    reasons over all of the non-changeset graphs in the store using the
    rules and for whatever statements it is able to infer adds them to
    the graphs and creates a new, (incremental) changeset.

    The resulting store is then passed back up to the handler where
    it should go out the outbound handlers and eventually be indexed.

    The configuration for this is slightly subtle, see the
    :ref:`configuring-inferencing` section of the configuration guide
    for how to do this.

    :param rule_modules: comma separated list of modules that export
        an :meth:`inference_rules` method
    """
    def __init__(self, rule_modules, foaf_mbox="fuxi-discussion@googlegroups.com", **kw):
        super(HandlerPlugin, self).__init__(**kw)
        self.rule_modules = map(strip, rule_modules.split(",")) if rule_modules else []
        self.busy = False
        ### FuXi, Agent of Change
        self.agent = Agent()
        self.agent.name("FuXi")
        self.agent.mbox(foaf_mbox)
        self.agent.nick(getuser())
        _s,_p,self.agent_hash = self.agent.one((self.agent.identifier, FOAF.mbox_sha1sum, None))

    def _get_network(self):
        if not hasattr(self, "__network__"):
            rstore, rgraph = SetupRuleStore()
            self.__network__ = ReteNetwork(rstore)
            log.info("startup %s" % self.__network__)
            for modname in self.rule_modules:
                __import__(modname, {}, {}, ["inference_rules"]
                           ).inference_rules(self.handler, self.__network__)
                log.info("rules from %s %s" % (modname, self.__network__))
        return self.__network__
    def _set_network(self, network):
        setattr(self, "__network__", network)
    def _del_network(self):
        if hasattr(self, "__network__"):
            delattr(self, "__network__")
    network = property(_get_network, _set_network, _del_network)

    def put(self, store):
        if self.busy:
            return
        self.busy = True

        inferredFacts = Graph()
        self.network.reset(inferredFacts)

        ## workaround for bug in reset()
        ## http://code.google.com/p/fuxi/issues/detail?id=17
        self.network.inferredFacts = inferredFacts

        if isinstance(store, _Graph):
            store = store.store

        def _get(graph):
            if isinstance(graph, basestring):
                identifier = graph
            else:
                identifier = graph.identifier
            for ctx in store.contexts():
                if ctx.identifier == identifier:
                    return ctx
        self.get = _get

        ### widen the set of statements to inference from
        ### any references, at one degree from the contents of
        ### the store, to external resources, add them to the
        ### store
        for s,p,o in ConjunctiveGraph(store).triples((None, None, None)):
            if isinstance(o, URIRef):
                o_graph = self.get(o)
                if not o_graph:
                    continue
                o_graph_copy = Graph(store=store, identifier=o_graph.identifier)
                for statement in o_graph.triples((None, None, None)):
                    o_graph_copy.add(statement)

        ### we don't reason on changesets, so make a list of changesets we have
        changesets = []
        facts = []
        for ctx in store.contexts():
            if ctx.exists((ctx.identifier, RDF.type, CS.ChangeSet)):
		log.debug("skip %s" % ctx.identifier)
                changesets.append(ctx)
            else:
                facts.append(ctx)

        ### retract previously inferred statements
        facts = [self.retract(g) for g in facts]

        ### feed data to the reasoner
        for fact_graph in facts:
            self.network.feedFactsToAdd(generateTokenSet(fact_graph))

        #log.debug("inferred\n%s" % inferredFacts.serialize(format="n3"))

        ### construct the change context

        change_ctx = self.handler.context(self.agent, "FuXi machine reasoning")
        for cs in changesets:
            ### for previous changeset, just preserve it in the change context
            ### store
            copy = Graph(store=change_ctx.store, identifier=cs.identifier)
            copy += cs

        for fact_graph in facts:
            ### add the relevant inferred facts
            copy = Graph(identifier=fact_graph.identifier)
            copy += fact_graph
            log.debug("facts %s" % fact_graph.identifier)
            for subj in fact_graph.distinct_subjects():
                inferred = inferredFacts.bnc((subj, None, None))
                log.debug("inferred %d facts about %s" % (len(inferred), subj))
                copy += inferred
                
            change_ctx.add(copy)

        ### save the changes
        ### this doesn't loop back through this function because
        ### at this point self.busy == True
        cs = change_ctx.commit()
        ### if we haven't been able to make any inferences, this changeset
        ### will be empty. but the preceding changeset will not, so we
        ### still need to send it to the indices
        if not cs:
            self.handler.put(change_ctx.store)

        #print cs.serialize(format="n3")

        ### get ready for the next message
        self.get = lambda x: None
        self.busy = False

        ### prevent further propagation of the original change bundle
        return True

    def retract(self, graph):
        """
        Retract statements previously inferred about graph by walking
        its change history and rolling back the most recent changeset
        that was made by this module
        """
        for cs in self.handler.history(graph.identifier):
            if isinstance(cs, list):
                log.warning("%s's change history branches before we get to inferred facts" % graph.identifier)
                break
            for creator in cs.objects(cs.identifier, DC.creator):
                if cs.exists((creator, FOAF.mbox_sha1sum, self.agent_hash)):
                    copy = Graph(identifier=graph.identifier)
                    copy += graph
                    cs.undo(copy)
                    ## preserve changeset linkage
                    copy.remove((copy.identifier, ORDF.changeSet, None))
                    [copy.add(change) for change in graph.triples((graph.identifier, ORDF.changeSet, None))]
                    return copy
        return graph



"""
Consuming from RabbitMQ -- :class:`RabbitHandler`
-------------------------------------------------
.. autoclass:: RabbitHandler
   :show-inheritance:

Publishing to RabbitMQ -- :class:`RabbitQueue`
----------------------------------------------
.. autoclass:: RabbitQueue
   :show-inheritance:
"""
__all__ = ['RabbitQueue', 'RabbitHandler']

from ordf.handler import Handler, HandlerPlugin
from ordf.graph import Graph, _Graph
from ordf.namespace import UUID
from ordf.term import URIRef
from carrot.messaging import Consumer, Publisher
from carrot.connection import BrokerConnection

from StringIO import StringIO
from threading import Thread
from logging import getLogger
from traceback import format_exc
from time import sleep

class _Rabbit(object):
    def __init__(self, storage=None, **kw):
        if isinstance(storage, basestring):
            kw.update(eval(storage))
        elif storage:
            kw.update(storage)
        self.__rabbit_kw__ = kw
    @property
    def connection(self):
        return BrokerConnection(**self.__rabbit_kw__)

class RabbitQueue(HandlerPlugin, _Rabbit):
    """
    Publisher of graps to the queueing system. This is a subclass of
    :class:`ordf.handler.HandlerPlugin` suitable to be registered as a
    reader. Any arguments are passed through to the connection 
    constructor :class:`carrot.connection.BrokerConnection` .

    .. automethod:: connect
    """
    def __init__(self, *av, **kw):
        super(RabbitQueue, self).__init__(*av, **kw)
        self.publisher = None
        self.connected = False

    def connect(self, **kw):
        """
        Parameters are interpreted as in :meth:`RabbitHandler.connect`
        though clearly machinery is set up for publishing data rather than
        starting threads to process incoming data. There is therefore no
        need for a **queue** keyword argument.

        :param exchange: highly recommended, defaults to "ordf"
        """
        kw.setdefault("exchange", "ordf")
        kw.setdefault("exchange_type", "fanout")
        kwa = kw.copy()
        kwa["exchange"] = kw["exchange"]
        kwa["exchange_type"] = "fanout"
        self.log = getLogger(self.__class__.__name__ + "/" + kwa["exchange"])
        self.log.info("connected")
        self.publisher = Publisher(connection=self.connection, **kwa)
        self.connected = True

    def put(self, graph):
        subgraphs = []
        if isinstance(graph, _Graph):
            contexts = [graph]
        else:
            contexts = graph.contexts()
        for ctx in contexts:
            self.log.debug("send %s" % (ctx.identifier,))
            subgraphs.append((ctx.identifier, unicode(ctx.serialize(), "utf-8")))
        self.publisher.send(subgraphs)

class RabbitHandler(Handler, _Rabbit):
    """
    This *Handler* listens to a RabbitMQ server for graphs and changesets.
    When it receives one, it is injected via :meth:`put` . As with the super-class
    read and write back-ends may be registered and the behaviour is identical

    .. automethod:: connect
    """
    def __init__(self, namespace=UUID, **kw):
        Handler.__init__(self, namespace=namespace)
        _Rabbit.__init__(self, **kw)

    def connect(self, **kw):
        """
        :synopsis: connect the handler to the RabbitMQ server.

        :param exchange: highly recommended, defaults to "ordf"
        :param queue: When specified forms the basis for the queue
            name. If the value were *"test"* then by default, two
            queues with names *test_graph* and *test_changesets* would
            be created and bound to the *ordf.graph* and *ordf.changesets*
            exchanges respectively.
        """
        self.__shutdown__ = False
        kwa = kw.copy()
        kwa.setdefault("exchange", "ordf")
        kwa.setdefault("exchange_type", "fanout")
        kwa.setdefault("queue", "ordf_test")

        def _loop(**kw):
            self.log = getLogger(self.__class__.__name__ + "/%(exchange)s/%(queue)s" % kw)
            self.log.info("consumer starting - exchange: %(exchange)s queue: %(queue)s" % kw)

            consumer = Consumer(connection=self.connection, **kw)

            def _recv(data, message):
                dummy = Graph("IOMemory")
                store = dummy.store
                try:
                    for ident, n3 in data:
                        graph = Graph(identifier=URIRef(ident), store=store)
                        self.log.info("recv %s" % (graph.identifier,))
                        graph.parse(StringIO(n3.encode("utf-8")))
                    self.put(store)
                except:
                    self.log.error(format_exc())
                        
            consumer.register_callback(_recv)

            while True:
                try:
                    msg = consumer.fetch(auto_ack=True, enable_callbacks=True)
                    if msg is None: ## queue empty
                        sleep(1)
                except Exception, e:
                    self.log.error("%s" % format_exc())
                finally:
                    if self.__shutdown__:
                        break

            self.log.info("consumer finishing")

        t = Thread(target=_loop, kwargs=kwa)
        t.start()

    def close(self):
        self.__shutdown__ = True




"""
.. autoclass:: Agent
   :show-inheritance:
.. autoclass:: Process
   :show-inheritance:
"""
__all__ = ["Agent", "Process"]

from ordf.term import BNode, Literal, URIRef, Node
from ordf.graph import Graph
from ordf.namespace import OPMV, ORDF, RDF, RDFS, TIME, FOAF
from ordf.utils import get_identifier
from ordf.vocab import foaf
from pkg_resources import get_distribution
from datetime import datetime
from socket import gethostname
from os import getpid

class Agent(foaf.Agent):
    """
    .. autoattribute:: __types__
    """
    __types__ = [OPMV["Agent"]]
    "* *opmv:Agent*"

    def __init__(self, *av, **kw):
        super(Agent, self).__init__(*av, **kw)
        self.add((self.identifier, RDF["type"], FOAF["Agent"]))

class Process(Graph):
    """
    .. autoattribute:: __types__

    .. automethod:: agent
    .. automethod:: use
    .. automethod:: result
    .. automethod:: add_distribution

    """
    __types__ = [OPMV["Process"]]
    """
    * *opmv:Process*
    """
    __distributions__ = {}
    @classmethod
    def add_distribution(cls, name):
        """
        Add a distribution (python package in the setuptools
        sense) by name to be recorded in provenance graph.

        This is a classmethod that should be called only once
        per distribution. The distribution "ordf" is already
        present.
        """
        try:
            dist = get_distribution(name)
        except DistributionNotFound:
            dist = None
        cls.__distributions__[name] = dist

    def __init__(self, *av, **kw):
        super(Process, self).__init__(*av, **kw)
        self.add((self.identifier, RDF.type, OPMV["Process"]))
        self.time = None
        self._finished = False
        self.add((self.identifier, ORDF["cmdline"], Literal(self.cmdline())))
        self.add((self.identifier, ORDF["hostname"], Literal(gethostname())))
        self.add((self.identifier, ORDF["pid"], Literal(getpid())))

        for name, dist in self.__distributions__.items():
            b = BNode()
            self.add((self.identifier, ORDF["version"], b))
            self.add((b, RDFS["label"], Literal(name)))
            if dist is not None:
                self.add((b, RDF["value"], Literal(dist.version)))

    def cmdline(self):
        from sys import argv
        args = []
        for a in argv:
            a = a.replace("'", "\\'")
            if " " in a or '"' in a:
                a = "'" + a + "'"
            args.append(a)
        return " ".join(args)
        
    def start(self):
        self.time = BNode()
        self.add((self.time, RDF.type, TIME["Interval"]))
        self.add((self.identifier, OPMV["wasPerformedAt"], self.time))

        start = BNode()
        self.add((self.time, TIME["hasBeginning"], start))
        self.add((start, RDF.type, TIME["Instant"]))
        self.add((start, TIME["inXSDDateTime"], Literal(datetime.utcnow())))

    def finish(self):
        if self._finished:
            return
        self._finished = True
        if self.time is None:
            self.time = BNode()
            self.add((self.time, RDF["type"], TIME["Instant"]))
            self.add((self.identifier, OPMV["wasPerformedAt"], self.time))
            self.add((self.time, TIME["inXSDDateTime"], Literal(datetime.utcnow())))
        else:
            end = BNode()
            self.add((self.time, TIME["hasEnd"], end))
            self.add((end, RDF.type, TIME["Instant"]))
            self.add((end, TIME["inXSDDateTime"], Literal(datetime.utcnow())))

    def clone(self):
        p = Process(identifier=self.identifier)
        p.remove((None, None, None))
        p.time = self.time
        p += self
        return p

    def use(self, what):
        """
        Add the given :class:`ordf.term.URIRef` or :class:`ordf.term.Graph` as
        a resource used by this process.
        """
        if isinstance(what, Node):
            ident = what
        else:
            ident = get_identifier(what)
        self.add((self.identifier, OPMV["used"], ident))

    def agent(self, agent):
        """
        Set the :class:`Agent` that is controlling this :class:`Process`

        :arg agent: a :class:`ordf.graph.Graph` or instance of :class:`Agent`
        """
        ident = get_identifier(agent)
        self.add((self.identifier, OPMV["wasControlledBy"], ident))
        self += agent

    def result(self, graph):
        """
        Cause linkages to this proces to be added to the given :class:`Graph`
        """
        self.finish()
        graph.add((graph.identifier, RDF["type"], OPMV["Artifact"]))
        graph.remove((graph.identifier, OPMV["wasGenereatedBy"], None))
        graph.add((graph.identifier, OPMV["wasGeneratedBy"], self.identifier))
        graph.remove((graph.identifier, OPMV["wasGenereatedAt"], None))
        graph.add((graph.identifier, OPMV["wasGeneratedAt"], self.time))
        graph += self
Process.add_distribution("ordf")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Overview of InfixOWL
--------------------

The core OWL abstract syntax is implemented based on this recipe:

    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/384122

Quoting the original InfixOWL docstrings,

    Python has the wonderful "in" operator and it would be nice
    to have additional infix operator like this. This recipe shows
    how (almost) arbitrary infix operators can be defined.

Some usage examples follow below.

>>> from ordf.namespace import register_ns
>>> exNs = Namespace("http://example.com/")
>>> register_ns("ex", exNs)
>>> g = Graph()

Now we have an empty graph, we can construct OWL classes in it
using the Python classes defined in this module

>>> a = Class(exNs.Opera,graph=g)

Now we can assert rdfs:subClassOf and owl:equivalentClass relationships 
(in the underlying graph) with other classes using the 'subClassOf' 
and 'equivalentClass' descriptors which can be set to a list
of objects for the corresponding predicates.

>>> a.subClassOf = [exNs.MusicalWork]

We can then access the rdfs:subClassOf relationships

>>> print list(a.subClassOf)
[Class: ex:MusicalWork ]

This can also be used against already populated graphs:

.. code-block:: python

    owlGraph = Graph().parse(OWL)
    namespace_manager.bind('owl', OWL, override=False)
    owlGraph.namespace_manager = namespace_manager
    list(Class(OWL.Class,graph=owlGraph).subClassOf)

    --> [Class: rdfs:Class ]

Operators are also available.  For instance we can add ex:Opera to the extension
of the ex:CreativeWork class via the '+=' operator

>>> a
Class: ex:Opera SubClassOf: ex:MusicalWork
>>> b = Class(exNs.CreativeWork,graph=g)
>>> b += a
>>> list(a.subClassOf)
[Class: ex:CreativeWork , Class: ex:MusicalWork ]

And we can then remove it from the extension as well

>>> b -= a
>>> a
Class: ex:Opera SubClassOf: ex:MusicalWork

Boolean class constructions can also  be created with Python operators
For example, The | operator can be used to construct a class consisting of a owl:unionOf 
the operands:

>>> c =  a | b | Class(exNs.Work,graph=g)
>>> c
( ex:Opera or ex:CreativeWork or ex:Work )

Boolean class expressions can also be operated as lists (using python list operators)

>>> del c[c.index(Class(exNs.Work,graph=g))]
>>> c
( ex:Opera or ex:CreativeWork )

The '&' operator can be used to construct class intersection:
      
>>> woman = Class(exNs.Female,graph=g) & Class(exNs.Human,graph=g)
>>> woman.identifier = exNs.Woman
>>> woman
( ex:Female and ex:Human )
>>> len(woman)
2

Enumerated classes can also be manipulated

>>> contList = [Class(exNs.Africa,graph=g),Class(exNs.NorthAmerica,graph=g)]
>>> EnumeratedClass(members=contList,graph=g)
{ ex:Africa ex:NorthAmerica }

owl:Restrictions can also be instanciated:

>>> Restriction(exNs.hasParent,graph=g,allValuesFrom=exNs.Human)
( ex:hasParent only ex:Human )

Restrictions can also be created using Manchester OWL syntax in 'colloquial' Python 

>>> exNs.hasParent |some| Class(exNs.Physician,graph=g)
( ex:hasParent some ex:Physician )
>>> Property(exNs.hasParent,graph=g) |max| Literal(1)
( ex:hasParent max 1 )

Datatype and Object Properties
------------------------------

.. autoclass:: predicate
    :show-inheritance:
.. autoclass:: object_predicate
    :show-inheritance:
.. autoclass:: cached_predicate
    :show-inheritance:
.. autoclass:: cached_object_predicate
    :show-inheritance:

Core classes
------------

.. autoclass:: Individual
.. autoclass:: AnnotatibleTerms
    :show-inheritance:
.. autoclass:: Ontology
    :show-inheritance:
.. autoclass:: Class
    :show-inheritance:
.. autoclass:: Property
    :show-inheritance:
"""
import os, itertools
from pprint import pprint

from ordf.collection import Collection
from ordf.graph import Graph
from ordf.namespace import bind_ns, Namespace, OWL, RDF, RDFS, XSD
from ordf.term import Node, BNode, Identifier, Literal, URIRef, Variable

from rdflib.util import first

##
## Data fixtures for ORDF
##
def rdf_data():
    from logging import getLogger
    log = getLogger(__name__)
    graph_uri = OWL[""][:-1]
    log.info("Fetching %s" % graph_uri)
    g = Graph(identifier=graph_uri).parse(graph_uri)
    log.info("Parsed %s (%d triples)" % (graph_uri, len(g)))
    yield g

def inference_rules(handler, network):
    from FuXi.Horn.HornRules import HornFromN3
    import os, pkg_resources
    rule_file = pkg_resources.resource_filename("ordf.vocab", os.path.join("n3", "owl-rules.n3"))
    rules = HornFromN3(rule_file)
    for rule in rules:
        network.buildNetworkFromClause(rule)
    
    owl = handler.get(OWL[""][:-1])
    if len(owl) == 0:
        for owl in rdf_data():
            pass
    dlp = network.setupDescriptionLogicProgramming(owl, addPDSemantics=True)
    return list(rules) + dlp

# attribute properties
class predicate(property):
    """
    This function provides facilities that look a bit like the properties
    on a `Django`_ or `SQLAlchemy`_ class.

    .. _Django: http://www.djangoproject.com/
    .. _SQLAlchemy: http://www.sqlalchemy.org/

    It is used when you make a subclass of :class:`Individual` or one
    of its subclasses so that you can put properties on. Here is an 
    example taken from the :mod:`ordf.vocab.foaf` module:

    >>> from ordf.namespace import FOAF
    >>> class Person(Class):
    ...     name = predicate(FOAF.name)
    ...
    >>> norman = Person()
    >>> norman.name = Literal("Norman Bethune")
    >>> [str(x) for x in norman.name]
    ['Norman Bethune']
    >>> del norman.name
    >>> len(list(norman.name))
    0

    The *set()* operation can take either a literal or a list/tuple
    of literals.
    """
    def __init__(self, term):
        self.term = term
        super(predicate, self).__init__(self._get, self._set, self._del)
    def _get(self, obj):
        return obj.graph.distinct_objects(obj.identifier, self.term)
    def _set(self, obj, value):
        if not value: return
        if not isinstance(value, list) and not isinstance(value, tuple):
            value = (value,)
        for x in value:
            if hasattr(x, "identifier") and isinstance(x.identifier, Node):
                x = x.identifier
            elif not isinstance(x, Node):
                x = Literal(x)
            obj.graph.add((obj.identifier, self.term, x))
    def _del(self, obj):
        obj.graph.remove((obj.identifier, self.term, None))

class object_predicate(predicate):
    """
    This function behaves similarly to :class:`predicate` except that
    rather than returning the term associated with the predicate it
    will attempt to return an instantiated Individual (or subclass
    thereof). In this way such a predicate will behave somewhat like
    a foreign key relationship as implemented in SQL ORMs.

    >>> from ordf.namespace import Namespace
    >>> EX = Namespace("http://example.org/")
    >>> class Country(Class):
    ...     pass
    ...
    >>> class Province(Class):
    ...     country = object_predicate(EX.country, Country)
    ...
    >>> canada = Country(EX.canada)
    >>> canada.label = "Canada"
    >>> quebec = Province(EX.quebec, graph=canada.graph)
    >>> quebec.label = "Québec"
    >>> quebec.country = canada
    >>> for country in quebec.country: print [str(x) for x in country.label]
    ['Canada']

    Try out some support for strings in place of class names for
    self-referential classes.

    >>> from ordf.namespace import FOAF
    >>> class Person(Class):
    ...     knows = object_predicate(FOAF.knows, "Person", globals())
    ...
    >>> bob = Person(EX.bob)
    >>> alice = Person(EX.alice)
    >>> bob.knows = alice
    >>> [type(x) for x in bob.knows]
    [<class 'ordf.vocab.owl.Person'>]
    """
    def __init__(self, term, implementation=None, globals=None):
        if implementation is None:
            self._cls = Individual
        if isinstance(implementation, basestring):
            self._cls = implementation
            self._globals = globals
        else:
            assert issubclass(implementation, Individual), "expected an Individual, got an %s" % implementation
            self._cls = implementation
        super(object_predicate, self).__init__(term)

    def _get(self, obj):
        if isinstance(self._cls, basestring):
            self._cls = self._globals[self._cls]
        for identifier in super(object_predicate, self)._get(obj):
            yield self._cls(identifier, graph=obj.graph, factoryGraph=obj.factoryGraph, 
                            skipClassMembership=True)

class cached_predicate(predicate):
    """
    This version of the :class:`predicate` function behaves the
    same way except that with *get()* operations the returned values
    are cached. Doing a *set()* or *del()* operation causes the cache
    to be cleared. Usage is essentially identical.

    >>> from ordf.namespace import DC
    >>> class Book(Class):
    ...    title = cached_predicate(DC.subject)
    ...
    >>> we = Book()
    >>> we.title = (Literal("We", lang="en"), Literal("Nous", lang="fr"))
    >>> [str(x) for x in we.title]
    ['We', 'Nous']
    >>> ### prove that the value is cached
    >>> we.graph.remove((we.identifier, None, None))
    >>> len(list(we.title))
    2
    >>> del we.title
    >>> len(list(we.title))
    0
    """
    def _cache(self, obj):
        if not hasattr(obj, "__predicate_cache__"):
            obj.__predicate_cache__ = {}
        return obj.__predicate_cache__
    def _get(self, obj):
        _cache = self._cache(obj)
        if self.term in _cache:
            return _cache[self.term]
        results = list(super(cached_predicate, self)._get(obj))
        _cache[self.term] = results
        return results
    def _set(self, obj, value):
        super(cached_predicate, self)._set(obj, value)
        _cache = self._cache(obj)
        if self.term in _cache: del _cache[self.term]
    def _del(self, obj):
        super(cached_predicate, self)._del(obj)
        _cache = self._cache(obj)
        if self.term in _cache: del _cache[self.term]

class cached_object_predicate(cached_predicate):
    """
    And the caching analogue to :class:`object_predicate`.

    >>> from ordf.namespace import Namespace
    >>> EX = Namespace("http://example.org/")
    >>> class Cat(Class):
    ...     pass
    >>> class Person(Class):
    ...     pet = cached_object_predicate(EX.pet, Cat)
    ...
    >>> bob = Person(EX.bob)
    >>> bob.label = "Bob"
    >>> max = Cat(EX.max, graph=bob.graph)
    >>> max.label = "Max"
    >>> bob.pet = max
    >>> for pet in bob.pet: print [str(x) for x in pet.label]
    ['Max']
    >>> # prove the data has been cached
    >>> bob.graph.remove((bob.identifier, None, None))
    >>> for pet in bob.pet: print [str(x) for x in pet.label]
    ['Max']
    >>> # prove that it is the *same* object
    >>> for pet in bob.pet:
    ...     assert id(pet) == id(max)
    ...
    """
    def __init__(self, term, implementation=None, globals=None):
        if implementation is None:
            self._cls = Individual
        elif isinstance(implementation, basestring):
            self._cls = implementation
            self._globals = globals
        else:
            assert issubclass(implementation, Individual), "expected an Individual, got an %s" % implementation
            self._cls = implementation
        super(cached_object_predicate, self).__init__(term)

    def _set(self, obj, value):
        if not isinstance(value, list) and not isinstance(value, tuple):
            value = [value]
        cached_values = self._get(obj)
        for x in value:
            if isinstance(x, URIRef) or isinstance(x, BNode):
                x = self._cls(value, graph=obj.graph)
            assert isinstance(x, Individual), "expected an Individual, got an %s" % x
            if x not in cached_values:
                cached_values.append(x)
                obj.graph.add((obj.identifier, self.term, x.identifier))

    def _get(self, obj):
        if isinstance(self._cls, basestring):
            self._cls = self._globals[self._cls]
        _cache = self._cache(obj)
        if self.term in _cache:
            return _cache[self.term]
        results = [
            self._cls(x, graph=obj.graph, factoryGraph=obj.factorGraph,
                      skipClassMembership=True)
            for x in super(cached_object_predicate, self)._get(obj)
            ]
        _cache[self.term] = results
        return results

# definition of an Infix operator class
# this recipe also works in jython
# calling sequence for the infix is either:
#  x |op| y
# or:
# x <<op>> y

class Infix(object):
    def __init__(self, function):
        self.function = function
    def __ror__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(other, x))
    def __or__(self, other):
        return self.function(other)
    def __rlshift__(self, other):
        return Infix(lambda x, self=self, other=other: self.function(other, x))
    def __rshift__(self, other):
        return self.function(other)
    def __call__(self, value1, value2):
        return self.function(value1, value2)

def generateQName(graph,uri):
    prefix,uri,localName = graph.compute_qname(classOrIdentifier(uri)) 
    return u':'.join([prefix,localName])    

def classOrTerm(thing):
    if isinstance(thing,Class):
        return thing.identifier
    else:
        assert isinstance(thing,(URIRef,BNode,Literal))
        return thing

def classOrIdentifier(thing):
    if isinstance(thing,(Property,Class)):
        return thing.identifier
    else:
        assert isinstance(thing,(URIRef,BNode)),"Expecting a Class, Property, URIRef, or BNode.. not a %s"%thing
        return thing

def propertyOrIdentifier(thing):
    if isinstance(thing,Property):
        return thing.identifier
    else:
        assert isinstance(thing,URIRef)
        return thing

def manchesterSyntax(thing,store,boolean=None,transientList=False):
    """
    Core serialization
    """
    assert thing is not None
    if boolean:
        if transientList:
            liveChildren=iter(thing)
            children = [manchesterSyntax(child,store) for child in thing ]
        else:
            liveChildren=iter(Collection(store,thing))
            children = [manchesterSyntax(child,store) for child in Collection(store,thing)]
        if boolean == OWL.intersectionOf:
            childList=[]
            named = []
            for child in liveChildren:
                if isinstance(child,URIRef):
                    named.append(child)
                else:
                    childList.append(child)
            if named:
                def castToQName(x):
                    prefix,uri,localName = store.compute_qname(x) 
                    return u':'.join([prefix,localName])
                
                if len(named) > 1:
                    prefix = '( '+ ' and '.join(map(castToQName,named)) + ' )'                
                else:
                    prefix = manchesterSyntax(named[0],store)
                if childList:
                    return prefix+ ' that '+' and '.join(
                             map(lambda x:manchesterSyntax(x,store),childList))
                else:
                    return prefix
            else:
                return '( '+ ' and '.join(children) + ' )'
        elif boolean == OWL.unionOf:
            return '( '+ ' or '.join(children) + ' )'
        elif boolean == OWL.oneOf:
            return '{ '+ ' '.join(children) +' }'
        else:            
            assert boolean == OWL.complementOf
    elif OWL.Restriction in store.objects(subject=thing, predicate=RDF.type):
        prop = list(store.objects(subject=thing, predicate=OWL.onProperty))[0]
        prefix,uri,localName = store.compute_qname(prop)
        propString = u':'.join([prefix,localName])
        for onlyClass in store.objects(subject=thing, predicate=OWL.allValuesFrom):
            return '( %s only %s )'%(propString,manchesterSyntax(onlyClass,store))
        for val in store.objects(subject=thing, predicate=OWL.hasValue):
            return '( %s value %s )'%(propString,manchesterSyntax(val,store))        
        for someClass in store.objects(subject=thing, predicate=OWL.someValuesFrom):    
            return '( %s some %s )'%(propString,manchesterSyntax(someClass,store))
        cardLookup = {OWL.maxCardinality:'max',OWL.minCardinality:'min',OWL.cardinality:'equals'}
        for s,p,o in store.triples_choices((thing,cardLookup.keys(),None)):            
            return '( %s %s %s )'%(propString,cardLookup[p],o.encode('utf-8'))
    compl = list(store.objects(subject=thing, predicate=OWL.complementOf)) 
    if compl:
        return '( not %s )'%(manchesterSyntax(compl[0],store))
    else:
        for boolProp,col in store.query("SELECT ?p ?bool WHERE { ?class a owl:Class; ?p ?bool . ?bool rdf:first ?foo }",
                                         initBindings={Variable("?class"):thing},
                                         initNs={ "owl": OWL, "rdf": RDF }):
            if not isinstance(thing,URIRef):                
                return manchesterSyntax(col,store,boolean=boolProp)
        try:
            prefix,uri,localName = store.compute_qname(thing) 
            qname = u':'.join([prefix,localName])
        except Exception,e:
            if isinstance(thing,BNode):
                return thing.n3()
            return "<"+thing+">"
            print list(store.objects(subject=thing,predicate=RDF.type))
            raise
            return '[]'#+thing._id.encode('utf-8')+'</em>'            
        label=first(Class(thing,graph=store).label)
        if label:
            return label.encode('utf-8')
        else:
            return qname.encode('utf-8')

def GetIdentifiedClasses(graph):
    for c in graph.subjects(predicate=RDF.type,object=OWL.Class):
        if isinstance(c,URIRef):
            yield Class(c)

def termDeletionDecorator(prop):
    def someFunc(func):
        func.property = prop
        return func
    return someFunc

class TermDeletionHelper:
    def __init__(self, prop):
        self.prop = prop
    def __call__(self, f):
        def _remover(inst):
            inst.graph.remove((inst.identifier,self.prop,None))
        return _remover

class Individual(object):
    """
    A typed individual, the base class of the InfixOWL classes.
    
    :param identifier:
        The identifier for this individual, a :class:`ordf.term.URIRef`

    :param graph:
        The graph which will provide the backing store for the individual.
        If *None* a new graph will be created.

    .. attribute:: identifier

        The identifier attribute is handled specially to make it 
        immutable.

    .. attribute:: graph

        The graph acting as a backing store is always accessible via
        this attribute on instances of :class:`Individual`

    .. attribute:: factoryGraph

        The usual practice is to set this equal to the :attr:`graph` 
        attribute, it is the source of data used by :meth:`serialize`

    .. attribute:: type

        Type is essentially the result of using the :func:`predicate` 
        function with *rdf:type*

    .. attribute:: cached_types

        A cached version of the :attr:`type` attribute.

    .. attribute:: sameAs

        Conceptually similar with *owl:sameAs* but special handling is
        required since it will return a full instance of a subclass of
        :class:`Individual`.

    .. automethod:: clearInDegree
    .. automethod:: clearOutDegree
    .. automethod:: delete
    .. automethod:: replace
    .. automethod:: serialize
    """
    def __init__(self, identifier=None, graph=None, factoryGraph=None, **kw):
        self.__identifier = identifier is not None and identifier or BNode()
        if factoryGraph is not None:
            self.factoryGraph = factoryGraph
        if graph is None:
            self.graph = self.factoryGraph
        else:
            self.graph = graph    
        self.qname = None
        if not isinstance(self.identifier,BNode):
            try:
                prefix,uri,localName = self.graph.compute_qname(self.identifier) 
                self.qname = u':'.join([prefix,localName])
            except:
                pass
    
    factoryGraph = Graph()
    def serialize(self, graph):
        """
        Take terms related to this individual using a blank node
        closure and add them to the provided graph.
        """
        graph += self.factoryGraph.bnc((self.identifier,None,None))

    def clearInDegree(self):
        """
        Remove references to this individual as an object in the
        backing store.
        """
        self.graph.remove((None,None,self.identifier))    

    def clearOutDegree(self):
        """
        Remove all statements to this individual as a subject in the
        backing store. Note that this only removes the statements 
        themselves, not the blank node closure so there is a chance
        that this will cause orphaned blank nodes to remain in the 
        graph.
        """
        self.graph.remove((self.identifier,None,None))    

    def delete(self):
        """
        Delete the individual from the graph, clearing the in and
        out degrees.
        """
        self.clearInDegree()
        self.clearOutDegree()
        
    def replace(self,other):
        """
        Replace the individual in the graph with the given other,
        causing all triples that refer to it to be changed and then
        delete the individual.
        """
        for s,p,o in self.graph.triples((None,None,self.identifier)):
            self.graph.add((s,p,classOrIdentifier(other)))
        self.delete()
    
    def _get_identifier(self):
        return self.__identifier
    def _set_identifier(self, i):
        assert i
        if i != self.__identifier:
            oldStmtsOut = [(p,o) for s,p,o in self.graph.triples((self.__identifier,None,None))]
            oldStmtsIn  = [(s,p) for s,p,o in self.graph.triples((None,None,self.__identifier))]
            for p1,o1 in oldStmtsOut:                
                self.graph.remove((self.__identifier,p1,o1))
            for s1,p1 in oldStmtsIn:                
                self.graph.remove((s1,p1,self.__identifier))
            self.__identifier = i
            self.graph.addN([(i,p1,o1,self.graph) for p1,o1 in oldStmtsOut])
            self.graph.addN([(s1,p1,i,self.graph) for s1,p1 in oldStmtsIn])
        if not isinstance(i,BNode):
            try:
                prefix,uri,localName = self.graph.compute_qname(i) 
                self.qname = u':'.join([prefix,localName])
            except:
                pass

    def __eq__(self, other):
        assert isinstance(other,Individual), repr(other)
        return self.identifier == other.identifier
            
    identifier = property(_get_identifier, _set_identifier)
    type = predicate(RDF.type)
    cached_types = cached_predicate(RDF.type)
    sameAs = predicate(OWL.sameAs)
    
class AnnotatibleTerms(Individual):
    """
    Terms in an OWL ontology with rdfs:label and rdfs:comment

    .. attribute:: isDefinedBy

        The result of calling the :func:`predicate` function on
        *rdfs:isDefinedBy*

    .. attribute:: label

        Similarly for *rdfs:label*

    .. attribute:: comment

        Similarly for *rdfs:comment*

    .. attribute:: seeAlso

        Similarly for *rdfs:seeAlso*
    """
    label = predicate(RDFS.label)
    comment = predicate(RDFS.comment)
    seeAlso = predicate(RDFS.seeAlso)
    isDefinedBy = predicate(RDFS.isDefinedBy)


class Ontology(AnnotatibleTerms):
    """
    The owl ontology metadata

    .. attribute:: versionInfo

        The result of calling :func:`predicate` on *owl:versionInfo*

    .. attribute:: imports

        Ditto with *owl:imports*
    """
    def __init__(self, identifier=None, imports=None, comment=None,graph=None):
        super(Ontology, self).__init__(identifier,graph)
        self.imports = imports and imports or []
        self.comment = comment and comment or []
        if (self.identifier,RDF.type,OWL.Ontology) not in self.graph:
            self.type = OWL.Ontology
    versionInfo = predicate(OWL.versionInfo)
    imports = predicate(OWL.imports)

def AllClasses(graph):
    prevClasses=set()
    for c in graph.subjects(predicate=RDF.type,object=OWL.Class):
        if c not in prevClasses:
            prevClasses.add(c)
            yield Class(c)            

def AllProperties(graph):
    prevProps=set()
    for s,p,o in graph.triples_choices(
               (None,RDF.type,[OWL.Symmetric,
                               OWL.FunctionalProperty,
                               OWL.InverseFunctionalProperty,
                               OWL.TransitiveProperty,
                               OWL.DatatypeProperty,
                               OWL.ObjectProperty])):
        if o in [OWL.Symmetric,
                 OWL.InverseFunctionalProperty,
                 OWL.TransitiveProperty,
                 OWL.ObjectProperty]:
            bType=OWL.ObjectProperty
        else:
            bType=OWL.DatatypeProperty
        if s not in prevProps:
            prevProps.add(s)
            yield Property(s,
                           graph=graph,
                           baseType=bType)            
    
class ClassNamespaceFactory(Namespace):
    def term(self, name):
        return Class(URIRef(self + name))

    def __getitem__(self, key, default=None):
        return self.term(key)

    def __getattr__(self, name):
        if name.startswith("__"): # ignore any special Python names!
            raise AttributeError
        else:
            return self.term(name)    
    
def DeepClassClear(classToPrune):
    """
    Recursively clear the given class, continuing
    where any related class is an anonymous class
    
    >>> g = Graph()    
    >>> EX = Namespace("http://example.com/")
    >>> bind_ns(g, { "ex": EX })
    >>> Individual.factoryGraph = g
    >>> classB = Class(EX.B)
    >>> classC = Class(EX.C)
    >>> classD = Class(EX.D)
    >>> classE = Class(EX.E)
    >>> classF = Class(EX.F)
    >>> anonClass = EX.someProp|some|classD
    >>> classF += anonClass
    >>> list(anonClass.subClassOf)
    [Class: ex:F ]
    >>> classA = classE | classF | anonClass    
    >>> classB += classA
    >>> classA.equivalentClass = [Class()]
    >>> classB.subClassOf = [EX.someProp|some|classC]
    >>> classA
    ( ex:E or ex:F or ( ex:someProp some ex:D ) )
    >>> DeepClassClear(classA)
    >>> classA
    (  )
    >>> list(anonClass.subClassOf)
    []
    >>> classB
    Class: ex:B SubClassOf: ( ex:someProp some ex:C )
    
    >>> otherClass = classD | anonClass
    >>> otherClass
    ( ex:D or ( ex:someProp some ex:D ) )
    >>> DeepClassClear(otherClass)
    >>> otherClass
    (  )
    >>> otherClass.delete()
    >>> list(g.triples((otherClass.identifier,None,None)))
    []
    """
    def deepClearIfBNode(_class):
        if isinstance(classOrIdentifier(_class),BNode):
            DeepClassClear(_class)
    classToPrune=CastClass(classToPrune,Individual.factoryGraph)
    for c in classToPrune.subClassOf:
        deepClearIfBNode(c)
    classToPrune.graph.remove((classToPrune.identifier,RDFS.subClassOf,None))
    for c in classToPrune.equivalentClass:
        deepClearIfBNode(c)
    classToPrune.graph.remove((classToPrune.identifier,OWL.equivalentClass,None))
    inverseClass = classToPrune.complementOf
    if inverseClass:
        classToPrune.graph.remove((classToPrune.identifier,OWL.complementOf,None))
        deepClearIfBNode(inverseClass)
    if isinstance(classToPrune,BooleanClass):
        for c in classToPrune:
            deepClearIfBNode(c)
        classToPrune.clear()
        classToPrune.graph.remove((classToPrune.identifier,
                          classToPrune._operator,
                          None))    
        
class MalformedClass(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __repr__(self): 
        return self.msg       
    
def CastClass(c,graph=None):
    graph = graph is None and c.factoryGraph or graph
    for kind in graph.objects(subject=classOrIdentifier(c),
                              predicate=RDF.type):
        if kind == OWL.Restriction:
            kwArgs = {'identifier':classOrIdentifier(c),
                      'graph'     :graph}
            for s,p,o in graph.triples((classOrIdentifier(c),
                                        None,
                                        None)):
                if p != RDF.type:
                    if p == OWL.onProperty:
                        kwArgs['onProperty'] = o
                    else:
                        if p not in Restriction.restrictionKinds:
                            continue
                        kwArgs[str(p.split(OWL)[-1])] = o
            if not set([str(i.split(OWL)[-1]) for i in Restriction.restrictionKinds]).intersection(kwArgs):
                raise MalformedClass("Malformed owl:Restriction")
            return Restriction(**kwArgs)
        else:
            for s,p,o in graph.triples_choices((classOrIdentifier(c),
                                                [OWL.intersectionOf,
                                                 OWL.unionOf,
                                                 OWL.oneOf],
                                                None)):
                if p == OWL.oneOf:
                    return EnumeratedClass(classOrIdentifier(c),graph=graph)
                else:
                    return BooleanClass(classOrIdentifier(c),operator=p,graph=graph)
            #assert (classOrIdentifier(c),RDF.type,OWL.Class) in graph
            return Class(classOrIdentifier(c),graph=graph,skipClassMembership=True)
    
class Class(AnnotatibleTerms):
    """
    'General form' for classes:
    
    The Manchester Syntax (supported in Protege) is used as the basis for the form 
    of this class
    
    See: http://owl-workshop.man.ac.uk/acceptedLong/submission_9.pdf::
    
        Class: classID {Annotation
            ( (SubClassOf: ClassExpression)
              | (EquivalentTo ClassExpression)
              | (DisjointWith ClassExpression)) }
    
    Appropriate excerpts from OWL Reference:
    
        ".. Subclass axioms provide us with partial definitions: they represent 
        necessary but not sufficient conditions for establishing class 
        membership of an individual."
     
        ".. A class axiom may contain (multiple) owl:equivalentClass statements"   
        
        "..A class axiom may also contain (multiple) owl:disjointWith statements.."
    
        "..An owl:complementOf property links a class to precisely one class 
        description."

    .. automethod:: get
    .. automethod:: get_or_create
    .. automethod:: query
    .. automethod:: filter
    """
    def _serialize(self,graph):
        for cl in self.subClassOf:
            CastClass(cl,self.graph).serialize(graph)
        for cl in self.equivalentClass:
            CastClass(cl,self.graph).serialize(graph)
        for cl in self.disjointWith:
            CastClass(cl,self.graph).serialize(graph)
        if self.complementOf:
            CastClass(self.complementOf,self.graph).serialize(graph)
        
    def serialize(self,graph):
        for fact in self.graph.triples((self.identifier,None,None)):
            graph.add(fact)
        self._serialize(graph)
    
    def __init__(self, identifier=None, graph=None, factoryGraph=None, factoryClass=Individual,
                 subClassOf=None,equivalentClass=None,
                 disjointWith=None,complementOf=None,
                 skipClassMembership=False,
                 comment=None):
        super(Class, self).__init__(identifier, graph=graph, factoryGraph=factoryGraph)
        if not skipClassMembership and (self.identifier,RDF.type,OWL.Class) not in self.graph and \
           (self.identifier,RDF.type,OWL.Restriction) not in self.graph:
            self.graph.add((self.identifier,RDF.type,OWL.Class))
        self.factoryClass    = factoryClass
        self.subClassOf      = subClassOf and subClassOf or [] 
        self.equivalentClass = equivalentClass and equivalentClass or []
        self.disjointWith    = disjointWith  and disjointWith or []
        if complementOf:
            self.complementOf    = complementOf
        self.comment = comment and comment or []
        
    def _get_extent(self,graph=None):
        for member in (graph is None and self.graph or graph).subjects(predicate=RDF.type,
                                          object=self.identifier):
            yield member
    def _set_extent(self,other):
        if not other:
            return
        for m in other:
            self.graph.add((classOrIdentifier(m),RDF.type,self.identifier))
    @TermDeletionHelper(RDF.type)
    def _del_type(self):            
        pass            
    extent = property(_get_extent, _set_extent, _del_type)            

    def _get_extentQuery(self):
        return (Variable('CLASS'),RDF.type,self.identifier)
    def _set_extentQuery(self,other): pass
    extentQuery = property(_get_extentQuery, _set_extentQuery)            

    def get(self, identifier):
        """
        Simple and naive individual retrieval from the factory graph.
        Raises a KeyError if no individual with type of the present
        class is found in the factory graph.

        >>> EX=Namespace("http://example.org/")
        >>> class Country(AnnotatibleTerms):
        ...     def __init__(self, *av, **kw):
        ...         super(Country, self).__init__(*av, **kw)
        ...         self.type = EX.Country
        ...
        >>> class CountryClass(Class):
        ...     def __init__(self, **kw):
        ...         super(CountryClass, self).__init__(EX.Country, factoryClass=Country)
        ...
        >>> country = CountryClass()
        >>> kenya = Country(EX.kenya)
        >>> kenya.label = "Kenya"
        >>> [str(x) for x in country.get(EX.kenya).label]
        ['Kenya']
        >>> country.get(EX.uganda)
        Traceback (most recent call last):
          raise KeyError("no individual %s with type %s" % (identifier, self.identifier))
        KeyError: u'no individual http://example.org/uganda with type http://example.org/Country'
        """
        if self.factoryGraph.exists((identifier, RDF.type, self.identifier)):
            return self.factoryClass(identifier, graph=self.factoryGraph)
        raise KeyError("no individual %s with type %s" % (identifier, self.identifier))

    def get_or_create(self, identifier):
        """
        Similar to the :meth:`get` except that it will not raise a key error
        and will create the individual requested.
        """
        return self.factoryClass(identifier, graph=self.factoryGraph)

    def query(self, *where):
        """
        Return a Telescope Select object that will query for individuals
        with the RDF type of the present class and additional where
        statements as provided.

        >>> from telescope import v
        >>> c = Class(URIRef("http://example.org/"))
        >>> print c.query((v.id, RDFS.label, Literal("Russia"))).compile()
        SELECT DISTINCT ?id
        WHERE { ?id <http://www.w3.org/1999/02/22-rdf-syntax-ns#type> <http://example.org/> . ?id <http://www.w3.org/2000/01/rdf-schema#label> "Russia" }
        """
        from telescope import Select, v
        q = Select([v.id], distinct=True).where(
            (v.id, RDF.type, self.identifier)
            ).where(*where)
        return q

    def filter(self, *where):
        """
        Return individuals that match the given query. The query will
        be constructed via :meth:`query`
        """
        q = self.query(*where)
        for identifier in self.factoryGraph.query(q.compile()):
            yield self.factoryClass(identifier, graph=self.factoryGraph)
            
    def __hash__(self):
        """
        >>> b=Class(OWL.Restriction)
        >>> c=Class(OWL.Restriction)
        >>> len(set([b,c]))
        1
        """
        return hash(self.identifier)

    def __eq__(self, other):
        assert isinstance(other,Class),repr(other)
        return self.identifier == other.identifier
    
    def __iadd__(self, other):
        assert isinstance(other,Class)
        other.subClassOf = [self]
        return self

    def __isub__(self, other):
        assert isinstance(other,Class)
        self.graph.remove((classOrIdentifier(other),RDFS.subClassOf,self.identifier))
        return self
    
    def __invert__(self):
        """
        Shorthand for Manchester syntax's not operator
        """
        return Class(complementOf=self)

    def __or__(self,other):
        """
        Construct an anonymous class description consisting of the union of this class and '
        other' and return it
        """
        return BooleanClass(operator=OWL.unionOf,members=[self,other],graph=self.graph)

    def __and__(self,other):
        """
        Construct an anonymous class description consisting of the intersection of this class and '
        other' and return it
        
        Chaining 3 intersections
        
        >>> g = Graph()    
        >>> exNs = Namespace("http://example.com/")
        >>> bind_ns(g, { "ex": exNs })
        >>> female      = Class(exNs.Female,graph=g)
        >>> human       = Class(exNs.Human,graph=g)
        >>> youngPerson = Class(exNs.YoungPerson,graph=g)
        >>> youngWoman = female & human & youngPerson
        >>> youngWoman
        ex:YoungPerson that ( ex:Female and ex:Human )
        >>> isinstance(youngWoman,BooleanClass)
        True
        >>> isinstance(youngWoman.identifier,BNode)
        True
        """
        return BooleanClass(operator=OWL.intersectionOf,members=[self,other],graph=self.graph)

    class _subClassOf(object_predicate):
        def __init__(self):
            class _xClass(Individual):
                def __new__(self, *av, **kw):
                    kwa=kw.copy()
                    kwa["skipClassMembership"] = True
                    return Class(*av, **kwa)
            super(self.__class__, self).__init__(RDFS.subClassOf, _xClass)
    subClassOf = _subClassOf()

    class _classPredicate(object_predicate):
        def __init__(self, term):
            class _xClass(Individual):
                def __new__(self, *av, **kw):
                    return Class(*av, **kw)
            super(self.__class__, self).__init__(term, _xClass)

    equivalentClass = _classPredicate(OWL.equivalentClass)
    disjointWith = _classPredicate(OWL.disjointWith)

    class _complement(object_predicate):
        def __init__(self):
            class _xClass(Individual):
                def __new__(self, *av, **kw):
                    return Class(*av, **kw)
            super(self.__class__, self).__init__(OWL.complementOf, _xClass)
        def _get(self, obj):
            comp = list(super(self.__class__, self)._get(obj))
            if not comp:
                return None
            elif len(comp) == 1:
                return Class(comp[0],graph=self.graph)
            else:
                raise Exception(len(comp))
    complementOf = _complement()

    def isPrimitive(self):
        if (self.identifier,RDF.type,OWL.Restriction) in self.graph:
            return False
        sc = list(self.subClassOf)
        ec = list(self.equivalentClass)
        for boolClass,p,rdfList in self.graph.triples_choices((self.identifier,
                                                               [OWL.intersectionOf,
                                                                OWL.unionOf],
                                                                None)):
            ec.append(manchesterSyntax(rdfList,self.graph,boolean=p))
        for e in ec:
            return False
        if self.complementOf:
            return False
        return True
    
    def subSumpteeIds(self):
        for s in self.graph.subjects(predicate=RDFS.subClassOf,object=self.identifier):
            yield s
        
#    def __iter__(self):
#        for s in self.graph.subjects(predicate=RDFS.subClassOf,object=self.identifier):
#            yield Class(s,skipClassMembership=True)
    
    def __repr__(self,full=False,normalization=True):
        """
        Returns the Manchester Syntax equivalent for this class
        """
        exprs = []
        sc = list(self.subClassOf)
        ec = list(self.equivalentClass)
        for boolClass,p,rdfList in self.graph.triples_choices((self.identifier,
                                                               [OWL.intersectionOf,
                                                                OWL.unionOf],
                                                                None)):
            ec.append(manchesterSyntax(rdfList,self.graph,boolean=p))
        dc = list(self.disjointWith)
        c  = self.complementOf
        if c:
            dc.append(c)
        klassKind = ''
        label = list(self.graph.objects(self.identifier,RDFS.label))
        label = label and '('+label[0]+')' or ''
        if sc:
            if full:
                scJoin = '\n                '
            else:
                scJoin = ', '
            necStatements = [
              isinstance(s,Class) and isinstance(self.identifier,BNode) and
                                      repr(CastClass(s,self.graph)) or
                                      #repr(BooleanClass(classOrIdentifier(s),
                                      #                  operator=None,
                                      #                  graph=self.graph)) or 
              manchesterSyntax(classOrIdentifier(s),self.graph) for s in sc]
            if necStatements:
                klassKind = "Primitive Type %s"%label
            exprs.append("SubClassOf: %s"%scJoin.join(necStatements))
            if full:
                exprs[-1]="\n    "+exprs[-1]
        if ec:
            nec_SuffStatements = [    
              isinstance(s,basestring) and s or 
              manchesterSyntax(classOrIdentifier(s),self.graph) for s in ec]
            if nec_SuffStatements:
                klassKind = "A Defined Class %s"%label
            exprs.append("EquivalentTo: %s"%', '.join(nec_SuffStatements))
            if full:
                exprs[-1]="\n    "+exprs[-1]
        if dc:
            exprs.append("DisjointWith %s\n"%'\n                 '.join([
              manchesterSyntax(classOrIdentifier(s),self.graph) for s in dc]))
            if full:
                exprs[-1]="\n    "+exprs[-1]
        descr = list(self.graph.objects(self.identifier,RDFS.comment))
        if full and normalization:
            klassDescr = klassKind and '\n    ## %s ##'%klassKind +\
            (descr and "\n    %s"%descr[0] or '') + ' . '.join(exprs) or ' . '.join(exprs)
        else:
            klassDescr = full and (descr and "\n    %s"%descr[0] or '') or '' + ' . '.join(exprs)
        return (isinstance(self.identifier,BNode) and "Some Class " or "Class: %s "%self.qname)+klassDescr

class OWLRDFListProxy(object):
    def __init__(self,rdfList,members=None):
        members = members and members or []
        if rdfList:
            self._rdfList = Collection(self.graph,rdfList[0])
            for member in members:
                if member not in self._rdfList:
                    self._rdfList.append(classOrIdentifier(member))
        else:
            self._rdfList = Collection(self.graph,BNode(),
                                       [classOrIdentifier(m) for m in members])
            self.graph.add((self.identifier,self._operator,self._rdfList.uri)) 
            
    def __eq__(self, other):
        """
        Equivalence of boolean class constructors is determined by equivalence of its
        members 
        """
        assert isinstance(other,Class),repr(other)+repr(type(other))
        if isinstance(other,BooleanClass):
            length = len(self)
            if length != len(other):
                return False
            else:
                for idx in range(length):
                    if self[idx] != other[idx]:
                        return False
                    return True
        else:
            return self.identifier == other.identifier

    #Redirect python list accessors to the underlying Collection instance
    def __len__(self):
        return len(self._rdfList)

    def index(self, item):
        return self._rdfList.index(classOrIdentifier(item))
    
    def __getitem__(self, key):
        return self._rdfList[key]

    def __setitem__(self, key, value):
        self._rdfList[key] = classOrIdentifier(value)
        
    def __delitem__(self, key):
        del self._rdfList[key]    
        
    def clear(self):
        self._rdfList.clear()    

    def __iter__(self):
        for item in self._rdfList:
            yield item

    def __contains__(self, item):
        for i in self._rdfList:
            if i == classOrIdentifier(item):
                return 1
        return 0
    
    def append(self, item):
        self._rdfList.append(item)    

    def __iadd__(self, other):
        self._rdfList.append(classOrIdentifier(other))
        return self

class EnumeratedClass(OWLRDFListProxy,Class):
    """
    Class for owl:oneOf forms:
    
    OWL Abstract Syntax is used
    
    axiom ::= 'EnumeratedClass(' classID ['Deprecated'] { annotation } { individualID } ')'
    
    >>> g = Graph()    
    >>> exNs = Namespace("http://example.com/")
    >>> bind_ns(g, { "ex": exNs })
    >>> Individual.factoryGraph = g
    >>> ogbujiBros = EnumeratedClass(exNs.ogbujicBros,
    ...                              members=[exNs.chime,
    ...                                       exNs.uche,
    ...                                       exNs.ejike])
    >>> ogbujiBros
    { ex:chime ex:uche ex:ejike }
    >>> print g.serialize(format='n3')
    <BLANKLINE>
    @prefix ex: <http://example.com/>.
    @prefix owl: <http://www.w3.org/2002/07/owl#>.
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    <BLANKLINE>
     ex:ogbujicBros a owl:Class;
         owl:oneOf ( ex:chime ex:uche ex:ejike ). 
    <BLANKLINE>
     ex:chime a owl:Class. 
    <BLANKLINE>
     ex:ejike a owl:Class. 
    <BLANKLINE>
     ex:uche a owl:Class. 
    """
    _operator = OWL.oneOf
    def isPrimitive(self):
        return False
    def __init__(self, identifier=None,members=None,graph=None):
        Class.__init__(self,identifier,graph = graph)
        members = members and members or []
        rdfList = list(self.graph.objects(predicate=OWL.oneOf,subject=self.identifier))
        OWLRDFListProxy.__init__(self, rdfList, members)
    def __repr__(self):
        """
        Returns the Manchester Syntax equivalent for this class
        """
        return manchesterSyntax(self._rdfList.uri,self.graph,boolean=self._operator)        
    
    def serialize(self,graph):
        clonedList = Collection(graph,BNode())
        for cl in self._rdfList:
            clonedList.append(cl)
            CastClass(cl, self.graph).serialize(graph)
        
        graph.add((self.identifier,self._operator,clonedList.uri))
        for s,p,o in self.graph.triples((self.identifier,None,None)):
            if p != self._operator:
                graph.add((s,p,o))
        self._serialize(graph)
    
BooleanPredicates = [OWL.intersectionOf,OWL.unionOf]

class BooleanClassExtentHelper:
    """
    >>> testGraph = Graph()
    >>> Individual.factoryGraph = testGraph
    >>> EX = Namespace("http://example.com/")
    >>> bind_ns(testGraph, { "ex": EX })
    >>> fire  = Class(EX.Fire)
    >>> water = Class(EX.Water) 
    >>> testClass = BooleanClass(members=[fire,water])
    >>> testClass2 = BooleanClass(operator=OWL.unionOf,members=[fire,water])
    >>> for c in BooleanClass.getIntersections():
    ...     print c
    ( ex:Fire and ex:Water )
    >>> for c in BooleanClass.getUnions():
    ...     print c
    ( ex:Fire or ex:Water )
    """    
    def __init__(self, operator):
        self.operator = operator
    def __call__(self, f):
        def _getExtent():
            for c in Individual.factoryGraph.subjects(self.operator):
                yield BooleanClass(c,operator=self.operator)            
        return _getExtent
    
class Callable:
    def __init__(self, anycallable):
        self.__call__ = anycallable    

class BooleanClass(OWLRDFListProxy,Class):
    """
    See: http://www.w3.org/TR/owl-ref/#Boolean
    
    owl:complementOf is an attribute of Class, however
    
    """
    @BooleanClassExtentHelper(OWL.intersectionOf)
    @Callable
    def getIntersections(): pass
    getIntersections = Callable(getIntersections)    

    @BooleanClassExtentHelper(OWL.unionOf)
    @Callable
    def getUnions(): pass
    getUnions = Callable(getUnions)    
            
    def __init__(self,identifier=None,operator=OWL.intersectionOf,
                 members=None,graph=None):
        if operator is None:
            props=[]
            for s,p,o in graph.triples_choices((identifier,
                                                [OWL.intersectionOf,
                                                 OWL.unionOf],
                                                 None)):
                props.append(p)
                operator = p
            assert len(props)==1,repr(props)
        Class.__init__(self,identifier,graph = graph)
        assert operator in [OWL.intersectionOf,OWL.unionOf], str(operator)
        self._operator = operator
        rdfList = list(self.graph.objects(predicate=operator,subject=self.identifier))
        assert not members or not rdfList,"This is a previous boolean class description!"+repr(Collection(self.graph,rdfList[0]).n3())        
        OWLRDFListProxy.__init__(self, rdfList, members)

    def serialize(self,graph):
        clonedList = Collection(graph,BNode())
        for cl in self._rdfList:
            clonedList.append(cl)
            CastClass(cl, self.graph).serialize(graph)
        
        graph.add((self.identifier,self._operator,clonedList.uri))
        
        for s,p,o in self.graph.triples((self.identifier,None,None)):
            if p != self._operator:
                graph.add((s,p,o))
        self._serialize(graph)

    def isPrimitive(self):
        return False

    def changeOperator(self,newOperator):
        """
        Converts a unionOf / intersectionOf class expression into one 
        that instead uses the given operator
        
        >>> testGraph = Graph()
        >>> Individual.factoryGraph = testGraph
        >>> EX = Namespace("http://example.com/")
        >>> bind_ns(testGraph, { "ex": EX })
        >>> fire  = Class(EX.Fire)
        >>> water = Class(EX.Water) 
        >>> testClass = BooleanClass(members=[fire,water])
        >>> testClass
        ( ex:Fire and ex:Water )
        >>> testClass.changeOperator(OWL.unionOf)
        >>> testClass
        ( ex:Fire or ex:Water )
        >>> try: testClass.changeOperator(OWL.unionOf)
        ... except Exception, e: print e
        The new operator is already being used!
        
        """
        assert newOperator != self._operator,"The new operator is already being used!"
        self.graph.remove((self.identifier,self._operator,self._rdfList.uri))
        self.graph.add((self.identifier,newOperator,self._rdfList.uri))
        self._operator = newOperator

    def __repr__(self):
        """
        Returns the Manchester Syntax equivalent for this class
        """
        return manchesterSyntax(self._rdfList.uri,self.graph,boolean=self._operator)

    def __or__(self,other):
        """
        Adds other to the list and returns self
        """
        assert self._operator == OWL.unionOf
        self._rdfList.append(classOrIdentifier(other))
        return self

def AllDifferent(members):
    """
    DisjointClasses(' description description { description } ')'
    
    """
    pass

class Restriction(Class):
    """
    restriction ::= 'restriction(' datavaluedPropertyID dataRestrictionComponent 
                                 { dataRestrictionComponent } ')'
                  | 'restriction(' individualvaluedPropertyID 
                      individualRestrictionComponent 
                      { individualRestrictionComponent } ')'    
    """
    
    restrictionKinds = [OWL.allValuesFrom,
                        OWL.someValuesFrom,
                        OWL.hasValue,
                        OWL.maxCardinality,
                        OWL.minCardinality]
    
    def __init__(self,
                 onProperty,
                 graph = Graph(),
                 allValuesFrom=None,
                 someValuesFrom=None,
                 value=None,
                 cardinality=None,
                 maxCardinality=None,
                 minCardinality=None,
                 identifier=None):
        super(Restriction, self).__init__(identifier,
                                          graph=graph,
                                          skipClassMembership=True)
        if (self.identifier,
            OWL.onProperty,
            propertyOrIdentifier(onProperty)) not in graph:
            graph.add((self.identifier,OWL.onProperty,propertyOrIdentifier(onProperty)))
        self.onProperty = onProperty
        restrTypes = [
                      (allValuesFrom,OWL.allValuesFrom ),
                      (someValuesFrom,OWL.someValuesFrom),
                      (value,OWL.hasValue),
                      (cardinality,OWL.cardinality),
                      (maxCardinality,OWL.maxCardinality),
                      (minCardinality,OWL.minCardinality)]
        validRestrProps = [(i,oTerm) for (i,oTerm) in restrTypes if i] 
        assert len(validRestrProps)
        restrictionRange,restrictionType = validRestrProps.pop()
        self.restrictionType = restrictionType
        if isinstance(restrictionRange,Identifier):
            self.restrictionRange = restrictionRange
        elif isinstance(restrictionRange,Class):
            self.restrictionRange = classOrIdentifier(restrictionRange)
        else:
            self.restrictionRange = first(self.graph.objects(self.identifier,
                                                             restrictionType))
        if (self.identifier,
            restrictionType,
            self.restrictionRange) not in self.graph:
            self.graph.add((self.identifier,restrictionType,self.restrictionRange))
        assert self.restrictionRange is not None,Class(self.identifier)
        if (self.identifier,RDF.type,OWL.Restriction) not in self.graph:
            self.graph.add((self.identifier,RDF.type,OWL.Restriction))
            self.graph.remove((self.identifier,RDF.type,OWL.Class))

    def serialize(self,graph):
        Property(self.onProperty,graph=self.graph).serialize(graph)
        for s,p,o in self.graph.triples((self.identifier,None,None)):
            graph.add((s,p,o))
            if p in [OWL.allValuesFrom,OWL.someValuesFrom]:
                CastClass(o, self.graph).serialize(graph)

    def isPrimitive(self):
        return False

    def __hash__(self):
        return hash((self.onProperty,self.restrictionRange))

    def __eq__(self, other):
        """
        Equivalence of restrictions is determined by equivalence of the property 
        in question and the restriction 'range'
        """
        assert isinstance(other,Class),repr(other)+repr(type(other))
        if isinstance(other,Restriction):
            return other.onProperty == self.onProperty and \
                   other.restrictionRange == self.restrictionRange
        else:
            return False

    def _get_onProperty(self):
        return list(self.graph.objects(subject=self.identifier,predicate=OWL.onProperty))[0]
    def _set_onProperty(self, prop):
        triple = (self.identifier,OWL.onProperty,propertyOrIdentifier(prop))
        if not prop:
            return
        elif triple in self.graph:
            return
        else:
            self.graph.set(triple)
            
    @TermDeletionHelper(OWL.onProperty)
    def _del_onProperty(self):            
        pass            
                        
    onProperty = property(_get_onProperty, _set_onProperty, _del_onProperty)

    def _get_allValuesFrom(self):
        for i in self.graph.objects(subject=self.identifier,predicate=OWL.allValuesFrom):
            return Class(i,graph=self.graph)
        return None
    def _set_allValuesFrom(self, other):
        triple = (self.identifier,OWL.allValuesFrom,classOrIdentifier(other))
        if not other:
            return
        elif triple in self.graph:
            return
        else:
            self.graph.set(triple)
            
    @TermDeletionHelper(OWL.allValuesFrom)
    def _del_allValuesFrom(self):            
        pass            
                        
    allValuesFrom = property(_get_allValuesFrom, _set_allValuesFrom, _del_allValuesFrom)

    def _get_someValuesFrom(self):
        for i in self.graph.objects(subject=self.identifier,predicate=OWL.someValuesFrom):
            return Class(i,graph=self.graph)
        return None
    def _set_someValuesFrom(self, other):
        triple = (self.identifier,OWL.someValuesFrom,classOrIdentifier(other))
        if not other:
            return
        elif triple in self.graph:
            return
        else:
            self.graph.set(triple)
            
    @TermDeletionHelper(OWL.someValuesFrom)
    def _del_someValuesFrom(self):            
        pass            
                        
    someValuesFrom = property(_get_someValuesFrom, _set_someValuesFrom, _del_someValuesFrom)

    def _get_hasValue(self):
        for i in self.graph.objects(subject=self.identifier,predicate=OWL.hasValue):
            return Class(i,graph=self.graph)
        return None
    def _set_hasValue(self, other):
        triple = (self.identifier,OWL.hasValue,classOrIdentifier(other))
        if not other:
            return
        elif triple in self.graph:
            return
        else:
            self.graph.set(triple)
            
    @TermDeletionHelper(OWL.hasValue)
    def _del_hasValue(self):            
        pass            
                        
    hasValue = property(_get_hasValue, _set_hasValue, _del_hasValue)

    def _get_cardinality(self):
        for i in self.graph.objects(subject=self.identifier,predicate=OWL.cardinality):
            return Class(i,graph=self.graph)
        return None
    def _set_cardinality(self, other):
        triple = (self.identifier,OWL.cardinality,classOrIdentifier(other))
        if not other:
            return
        elif triple in self.graph:
            return
        else:
            self.graph.set(triple)
            
    @TermDeletionHelper(OWL.cardinality)
    def _del_cardinality(self):            
        pass            
                        
    cardinality = property(_get_cardinality, _set_cardinality, _del_cardinality)

    def _get_maxCardinality(self):
        for i in self.graph.objects(subject=self.identifier,predicate=OWL.maxCardinality):
            return Class(i,graph=self.graph)
        return None
    def _set_maxCardinality(self, other):
        triple = (self.identifier,OWL.maxCardinality,classOrIdentifier(other))
        if not other:
            return
        elif triple in self.graph:
            return
        else:
            self.graph.set(triple)
            
    @TermDeletionHelper(OWL.maxCardinality)
    def _del_maxCardinality(self):            
        pass            
                        
    maxCardinality = property(_get_maxCardinality, _set_maxCardinality, _del_maxCardinality)

    def _get_minCardinality(self):
        for i in self.graph.objects(subject=self.identifier,predicate=OWL.minCardinality):
            return Class(i,graph=self.graph)
        return None
    def _set_minCardinality(self, other):
        triple = (self.identifier,OWL.minCardinality,classOrIdentifier(other))
        if not other:
            return
        elif triple in self.graph:
            return
        else:
            self.graph.set(triple)
            
    @TermDeletionHelper(OWL.minCardinality)
    def _del_minCardinality(self):            
        pass            
                        
    minCardinality = property(_get_minCardinality, _set_minCardinality, _del_minCardinality)

    def restrictionKind(self):
        for p in self.graph.triple_choices((self.identifier,
                                            self.restrictionKinds,
                                            None)):
            return p.split(OWL)[-1]
        raise
            
    def __repr__(self):
        """
        Returns the Manchester Syntax equivalent for this restriction
        """
        return manchesterSyntax(self.identifier,self.graph)

### Infix Operators ###

some     = Infix(lambda prop,_class: Restriction(prop,graph=_class.graph,someValuesFrom=_class))
only     = Infix(lambda prop,_class: Restriction(prop,graph=_class.graph,allValuesFrom=_class))
max      = Infix(lambda prop,_class: Restriction(prop,graph=prop.graph,maxCardinality=_class))
min      = Infix(lambda prop,_class: Restriction(prop,graph=prop.graph,minCardinality=_class))
exactly  = Infix(lambda prop,_class: Restriction(prop,graph=prop.graph,cardinality=_class))
value    = Infix(lambda prop,_class: Restriction(prop,graph=prop.graph,value=_class))

PropertyAbstractSyntax=\
"""
%s( %s { %s } 
%s
{ 'super(' datavaluedPropertyID ')'} ['Functional']
{ domain( %s ) } { range( %s ) } )"""

class Property(AnnotatibleTerms):
    """
    What is a property?::

        axiom ::= 'DatatypeProperty(' datavaluedPropertyID ['Deprecated'] { annotation } 
                    { 'super(' datavaluedPropertyID ')'} ['Functional']
                    { 'domain(' description ')' } { 'range(' dataRange ')' } ')'
                | 'ObjectProperty(' individualvaluedPropertyID ['Deprecated'] { annotation } 
                    { 'super(' individualvaluedPropertyID ')' }
                    [ 'inverseOf(' individualvaluedPropertyID ')' ] [ 'Symmetric' ] 
                    [ 'Functional' | 'InverseFunctional' | 'Functional' 'InverseFunctional' |
                      'Transitive' ]
                    { 'domain(' description ')' } { 'range(' description ')' } ')    
    """
    def __init__(self,identifier=None,graph = None,baseType=OWL.ObjectProperty,
                      subPropertyOf=None,domain=None,range=None,inverseOf=None,
                      otherType=None,equivalentProperty=None,comment=None):
        super(Property, self).__init__(identifier,graph)
        assert not isinstance(self.identifier,BNode)
        if (self.identifier,RDF.type,baseType) not in self.graph:
            self.graph.add((self.identifier,RDF.type,baseType))
        self._baseType=baseType
        self.subPropertyOf = subPropertyOf
        self.inverseOf     = inverseOf
        self.domain        = domain
        self.range         = range
        self.comment = comment and comment or []

    def serialize(self,graph):
        for fact in self.graph.triples((self.identifier,None,None)):
            graph.add(fact)
        for p in itertools.chain(self.subPropertyOf,
                                 self.inverseOf):
            p.serialize(graph)
        for c in itertools.chain(self.domain,
                                 self.range):
            CastClass(c,self.graph).serialize(graph)

    def _get_extent(self,graph=None):
        for triple in (graph is None and 
                       self.graph or graph).triples((None,self.identifier,None)):
            yield triple
    def _set_extent(self,other):
        if not other:
            return
        for subj,obj in other:
            self.graph.add((subj,self.identifier,obj))
            
    extent = property(_get_extent, _set_extent)            

    def __repr__(self):
        rt=[]
        if OWL.ObjectProperty in self.type:
            rt.append(u'ObjectProperty( %s annotation(%s)'\
                       %(self.qname,first(self.comment) and first(self.comment) or ''))
            if first(self.inverseOf):
                twoLinkInverse=first(first(self.inverseOf).inverseOf)
                if twoLinkInverse and twoLinkInverse.identifier == self.identifier:
                    inverseRepr=first(self.inverseOf).qname
                else:
                    inverseRepr=repr(first(self.inverseOf))
                rt.append(u"  inverseOf( %s )%s"%(inverseRepr,
                            OWL.Symmetric in self.type and u' Symmetric' or u''))
            for s,p,roleType in self.graph.triples_choices((self.identifier,
                                                            RDF.type,
                                                            [OWL.Functional,
                                                             OWL.InverseFunctionalProperty,
                                                             OWL.Transitive])):
                rt.append(unicode(roleType.split(OWL)[-1]))
        else:
            rt.append('DatatypeProperty( %s %s'\
                       %(self.qname,first(self.comment) and first(self.comment) or ''))            
            for s,p,roleType in self.graph.triples((self.identifier,
                                                    RDF.type,
                                                    OWL.Functional)):
                rt.append(u'   Functional')
        def canonicalName(term,g):
            normalizedName=classOrIdentifier(term)
            if isinstance(normalizedName,BNode):
                return term
            elif normalizedName.startswith(XSD):
                return unicode(term)
            elif first(g.triples_choices((
                      normalizedName,
                      [OWL.unionOf,
                       OWL.intersectionOf],None))):
                return repr(term)
            else:
                return unicode(term.qname)
        rt.append(u' '.join([u"   super( %s )"%canonicalName(superP,self.graph) 
                              for superP in self.subPropertyOf]))                        
        rt.append(u' '.join([u"   domain( %s )"% canonicalName(domain,self.graph) 
                              for domain in self.domain]))
        rt.append(u' '.join([u"   range( %s )"%canonicalName(range,self.graph) 
                              for range in self.range]))
        rt=u'\n'.join([expr for expr in rt if expr])
        rt+=u'\n)'
        return unicode(rt).encode('utf-8')
                    
    def _get_subPropertyOf(self):
        for anc in self.graph.objects(subject=self.identifier,predicate=RDFS.subPropertyOf):
            yield Property(anc,graph=self.graph)
    def _set_subPropertyOf(self, other):
        if not other:
            return        
        for sP in other:
            self.graph.add((self.identifier,RDFS.subPropertyOf,classOrIdentifier(sP)))
            
    @TermDeletionHelper(RDFS.subPropertyOf)
    def _del_subPropertyOf(self):            
        pass            
                        
    subPropertyOf = property(_get_subPropertyOf, _set_subPropertyOf, _del_subPropertyOf)

    def _get_inverseOf(self):
        for anc in self.graph.objects(subject=self.identifier,predicate=OWL.inverseOf):
            yield Property(anc,graph=self.graph)
    def _set_inverseOf(self, other):
        if not other:
            return        
        self.graph.add((self.identifier,OWL.inverseOf,classOrIdentifier(other)))
        
    @TermDeletionHelper(OWL.inverseOf)
    def _del_inverseOf(self):            
        pass            
                
    inverseOf = property(_get_inverseOf, _set_inverseOf, _del_inverseOf)

    def _get_domain(self):
        for dom in self.graph.objects(subject=self.identifier,predicate=RDFS.domain):
            yield Class(dom,graph=self.graph)
    def _set_domain(self, other):
        if not other:
            return
        if isinstance(other,(Individual,Identifier)):
            self.graph.add((self.identifier,RDFS.domain,classOrIdentifier(other)))
        else:
            for dom in other:
                self.graph.add((self.identifier,RDFS.domain,classOrIdentifier(dom)))
                
    @TermDeletionHelper(RDFS.domain)
    def _del_domain(self):            
        pass            
                                
    domain = property(_get_domain, _set_domain, _del_domain)

    def _get_range(self):
        for ran in self.graph.objects(subject=self.identifier,predicate=RDFS.range):
            yield Class(ran,graph=self.graph)
    def _set_range(self, ranges):
        if not ranges:
            return
        if isinstance(ranges,(Individual,Identifier)):
            self.graph.add((self.identifier,RDFS.range,classOrIdentifier(ranges)))
        else:        
            for range in ranges:
                self.graph.add((self.identifier,RDFS.range,classOrIdentifier(range)))
                
    @TermDeletionHelper(RDFS.range)
    def _del_range(self):            
        pass            
                                
    range = property(_get_range, _set_range, _del_range)

        
def test():
    import doctest
    doctest.testmod()

if __name__ == '__main__':
    test()



"""
Test rdflib storage of models
"""
import tempfile
import shutil

from ordf.tests import our_test_graph, identifier
from ordf.handler import Handler
from ordf.handler.rdf import RDFLib


class TestRdflibStorage:
    @classmethod
    def setup_class(self):
        self.tmpdir = tempfile.mkdtemp()
        self.rdf = RDFLib(self.tmpdir, store="Sleepycat")

    @classmethod
    def teardown_class(self):
        shutil.rmtree(self.tmpdir)

    def test_01_create(self):
        self.rdf[identifier] = our_test_graph()

    def test_02_read(self):
        g = self.rdf[identifier]

    def test_03_delete(self):
        del self.rdf[identifier]

    def test_04_graphops(self):
        self.rdf.put(graph=our_test_graph())
        self.rdf.get(our_test_graph())
        self.rdf.remove(our_test_graph())

    def test_05_handler(self):
        handler = Handler()
        handler.register_reader(self.rdf)
        handler.register_writer(self.rdf)
        ctx = handler.context("rdflib_test", "rdflib_test")
        ctx.add(our_test_graph())
        ctx.commit()

        for cs in handler.history(identifier):
            self.rdf.remove(cs)
        self.rdf.remove(our_test_graph())




"""
Miscellaneous Utilities
=======================

.. autofunction:: get_identifier
.. autofunction:: uuid
"""

__all__ = ["uuid", "get_identifier"]

from uuid import uuid1, uuid3, NAMESPACE_URL
from ordf.graph import _Graph
from ordf.term import URIRef

def uuid(uri):
    """
    Convenience method - if the argument starts with *urn:uuid:*
    does nothing. Otherwise constructs a *urn:uuid:* form URI from
    the given argument. Uses the :func:`uuid.uuid3` function from the
    standard Python library
    """
    identifier = get_identifier(uri)
    if identifier.startswith("urn:uuid:"):
        return identifier
    return u"urn:uuid:" + unicode(uuid3(NAMESPACE_URL, identifier.encode("ascii")))

def get_identifier(g):
    """
    Returns a :class:`~rdflib.term.URIRef` for the argument which may be an 
    instance of :class:`~rdflib.graph.Graph` or a string. If it is already a
    :class:`~rdflib.term.URIRef` (which itself is a subclass of unicode string)
    it simply returns it.
    """
    if isinstance(g, _Graph):
        return g.identifier
    if isinstance(g, URIRef):
        return g
    return URIRef(g)




from ordf.graph import Graph
from ordf.term import URIRef
import sys, os

identifier = URIRef("http://example.org/ordf/")
def our_test_graph():
    return fixture("test.rdf", identifier)

def card_graph():
    return fixture("card.rdf", URIRef("http://example.org/swh.rdf"))

def fixture(filename, identifier=None):
    curdir = os.path.dirname(__file__)
    datafile = os.path.join(curdir, "data", filename)
    g = Graph(identifier=identifier)
    g.parse(datafile, format="n3" if filename.endswith(".n3") else "xml")
    return g

def nosep(s):
    sys.stdout.write("%s... " % s)
    sys.stdout.flush()



"""
This module brings OpenCyc piecemeal into ORDF/RDFLib applications.

.. autoclass:: Concept
   :show-inheritance:
.. autofunction:: rdf_data
.. _OpenCyc: http://sw.opencyc.org/
"""

from ordf.graph import ConjunctiveGraph, Graph
from ordf.namespace import register_ns, Namespace, RDF, RDFS, OWL
from ordf.term import BNode, Literal
from urllib import urlencode
from urllib2 import urlopen, HTTPError

CYC = Namespace("http://sw.opencyc.org/concept/")
CYC_ANNOT = Namespace("http://sw.cyc.com/CycAnnotations_v1#")
register_ns("cyc", CYC)
register_ns("cycAnnot", CYC_ANNOT)

from logging import getLogger
log = getLogger(__name__)

def rdf_data():
    """
    Data fixture for OpenCyc. Starts with the top level predicate
    and recurses through all 'owl:DatatypeProperty', 'owl:ObjectProperty',
    predicates and classes present in the returned RDF either as
    subjects or appearing in 'rdfs:domain' or 'rdfs:range'. In this
    way builds up a basic ontology that can be used for reasoning.

    This function may take some time to complete as at the time of
    writing it will yield some 520 distinct graphs.
    """
    g = Graph(identifier=CYC_ANNOT[""])
    g.parse(CYC_ANNOT[""])
    yield g

    seen = set()
    def _f(ident):
        if ident.startswith(CYC[""]) and ident not in seen:
            seen.add(ident)
            x = Graph(identifier=ident)
            log.info("fetching %s" % ident)
            try:
                _g = Graph().parse(ident)
            except HTTPError, e:
                log.error("error fetching %s: %s" % (ident, e))
                return
            x += _g.bnc((ident, None, None))
            yield x
            for s in _g.distinct_subjects(RDF["type"], RDF["Property"]):
                for x in _f(s): yield x
            for s in _g.distinct_subjects(RDF["type"], OWL["AnnotationProperty"]):
                for x in _f(s): yield x
            for s in _g.distinct_subjects(RDF["type"], OWL["DatatypeProperty"]):
                for x in _f(s): yield x
            for s in _g.distinct_subjects(RDF["type"], OWL["ObjectProperty"]):
                for x in _f(s): yield x
            for s in _g.distinct_subjects(RDF["type"], OWL["Class"]):
                for x in _f(s): yield x
            for s in _g.distinct_subjects(RDFS["subClassOf"]):
                for x in _f(s): yield x
            for p in _g.distinct_predicates():
                for x in _f(p): yield x
            for o in _g.bnc((None, RDFS["domain"], None)).distinct_objects():
                for x in _f(o): yield x
            for o in _g.bnc((None, RDFS["range"], None)).distinct_objects():
                for x in _f(o): yield x
            for o in _g.bnc((None, RDFS["subClassOf"], None)).distinct_objects():
                for x in _f(o): yield x

    for g in _f(CYC["Mx4rvViA1pwpEbGdrcN5Y29ycA"]): ## top level predicate
        log.info("Seen %d distinct concepts" % len(seen))
        yield g

class Concept(Graph):
    """
    OpenCyc is very big. For most purposes we don't want to have to store
    the entire knowledge base in our local store, but for inferencing 
    purposes we often will want to store some of the relevant concepts.

    OpenCyc handily provides a search interface that gives results in
    XML, and we use this to retrieve a concept if we don't know it's URI
    beforehand.

    Initialisation of :class:`Concept` takes, as with any other :class:`Graph`
    an optional 'store' argument. If the concept in question does not
    exist in the store it will be fetched and added. The data that is
    returned by OpenCyc will typically include some other resources,
    these are filtered out. Only the blank node closure of the requested
    resource is added to the store.

    The :class:`Concept` class has methods for walking the ontology
    tree, documented below. These can be useful for adding relevant
    resources to the store in an automated way, but can be quite slow
    as they can make a potentially large number of HTTP requests.

    >>> lamb, = Concept.search("lamb", max=1)
    >>> print lamb.cycAnnot()
    (JuvenileFn Sheep)
    >>> for parent in lamb.parents():
    ...     print parent.cycAnnot()
    ...
    Sheep
    JuvenileAnimal
    >>>

    .. automethod:: search
    .. automethod:: parents
    .. automethod:: ancestors
    .. automethod:: cycAnnot
    """
    _not_found = set()
    def __init__(self, *av, **kw):
        super(Concept, self).__init__(*av, **kw)
        if self.one((None, None, None)) is None: 
            log.info("fetching %s" % self.identifier)
            try:
                g = Graph().parse(self.identifier)
                self += g.bnc((self.identifier, None, None))
            except HTTPError, e:
                log.error("error fetching %s: %s" % (self.identifier, e))
                self._not_found.add(self.identifier)

    _nss = { "cyc": "http://ws.opencyc.org/xsd/CycConcepts" }
    @classmethod
    def search(self, name, max=1, exact=True, store="IOMemory"):
        """
        The OpenCyc search interface is not documented anywhere obvious, but
        a very small amount of reverse engineering the JavaScript code on the
        website and analysing the XML given in returned is sufficient to
        implement this method.

        :param name: a text string to search on. This might be the name
            in English of the concept that is of interest
        :param max: maximum number of results to return
        :param exact: exact matches only
        :param store: the RDFLib Store to which any results should be
            added, returned graphs are initialised with this store. The
            default is the string '"IOMemory"'
        :return: an iterator over populated :class:`Concept` graphs for
            each of the search results
        """
        from Ft.Xml.Domlette import NonvalidatingReader
        params = {
            "max": max,
            "isExactMatch": "true" if exact else "false",
            "conceptDetails": "typical",
            "str": name
        }
        uri = "http://sw.opencyc.org//webservices/concept/find?" + urlencode(params)
        fp = urlopen(uri)
        result = NonvalidatingReader.parseStream(fp, uri=uri)
        fp.close()

        for concept in result.xpath("/cyc:concepts/cyc:concept", explicitNss=self._nss):
            extid = concept.xpath("string(cyc:externalId)", explicitNss=self._nss)
            ident = CYC[extid]
            yield Concept(store, identifier=ident)

    def cycAnnot(self):
        """
        Return the Cyc annotation or representation in the Cyc language
        of the current resource.
        """
        _s, _p, o = self.one((self.identifier, CYC_ANNOT["label"], None))
        return o

    def parents(self, restrict=False, seen=set()):
        """
        Walk one step up the class hierarchy by following 'rdfs:subClassOf' 
        links.

        :param restrict: boolean indicating whether parents returned should 
            be restricted to the OpenCyc namespace.
        :param seen: set of identifiers that have already been processed and
            are not therefore to be returned, in order to avoid needless 
            recursion
        :return: an iterator yielding :class:`Concept` for each of the parent
            concepts.
        """
        for cls in self.distinct_objects(self.identifier, RDFS["subClassOf"]):
            if cls in seen: continue
            if restrict and not cls.startswith(CYC[""]): continue
            yield Concept(self.store, identifier=cls)

    def ancestors(self, restrict=False, seen=set()):
        """
        Walk to the top of the class hierarchy recursively using :meth:`parents`.
        Parameters are as with that method.
        """
        for parent in self.parents(restrict=restrict, seen=seen):
            if parent.identifier in seen: continue
            seen.add(parent.identifier)
            yield parent
            for ancestor in parent.ancestors(restrict=restrict, seen=seen):
                if ancestor.identifier in seen: continue
                seen.add(ancestor.identifier)
                yield ancestor

if __name__ == '__main__':
    from logging import basicConfig, DEBUG
    basicConfig(level=DEBUG)

    import doctest
    doctest.testmod()

    for g in rdf_data():
        statement = g.one((g.identifier, RDFS["label"], None))
        if statement is not None:
            print statement[0], statement[2]



"""
.. autoclass:: Graph

"""
__all__ = ["Graph", "ConjunctiveGraph", "ReadOnlyGraphAggregate"]

try:
    from rdflib.graph import Graph as _Graph
    from rdflib.graph import ReadOnlyGraphAggregate as _ROGraph
    from rdflib.graph import ConjunctiveGraph as _CGraph
    __rdflib_version__ = 3
except ImportError:
    from rdflib.Graph import Graph as _Graph
    from rdflib.Graph import ReadOnlyGraphAggregate as _ROGraph
    from rdflib.Graph import ConjunctiveGraph as _CGraph
    __rdflib_version__ = 2

### This is here because ordf/__init__.py is for the namespace package
### it is not really conceivable that any of the functions affected by
### this data get run without using ordf.graph
__import__("ordf.namespace", globals(), locals(), ["_init_ns"])._init_ns()
__import__("ordf.serializer") ## force registration of rdflib plugins

from ordf.namespace import bind_ns, RDF, RDFS, ORDF, RDFG
from ordf.term import BNode, Literal, URIRef, Node

class _Common(object):
    def bnc(self, triple, *av, **kw):
        """
        Return the BNode closure(s) for triples that are matched
        by the given "triple". Any additional positional or keyword
        arguments are passed to the constructor for the new graph.
        """
        result = Graph(*av, **kw)
        log = __import__("logging").getLogger(__name__)
        for s,p,o in self.triples(triple):
            result.add((s,p,o))
            if isinstance(o, BNode) and (o, None, None) not in result:
                result += self.bnc((o, None, None))
        return result

    def replace(self, old, new, *av, **kw):
        """
        Return a graph where triple "old" is replaced with triple
        "new". Any additional positional or keyword arguments are
        passed to the constructor for the new graph.
        """
        result = Graph(*av, **kw)
        result += self
        result.remove(old)
        ns, np, no = new
        for s,p,o in self.triples(old):
            if ns: s = ns
            if np: p = np
            if no: o = no
            result.add((s,p,o))
        return result

    def one(self, triple):
        """
        Return one matching "triple" or "None"
        """
        for statement in self.triples(triple):
            return statement

    def exists(self, triple):
        """
        Return "True" if "triple" exists, "False" otherwise
        """
        statement = self.one(triple)
        if statement is not None:
            return True
        return False

    def _distinct(self, func, *av, **kw):
        results = {}
        for x in func(*av, **kw):
            results[x] = True
        return results.keys()

    def distinct_subjects(self, *av, **kw):
        """
        Return a *distinct* set of subjects. Arguments are 
        as for :meth:subjects
        """
        return self._distinct(self.subjects, *av, **kw)

    def distinct_predicates(self, *av, **kw):
        """
        Return a *distinct* set of predicates. Arguments are 
        as for :meth:predicates
        """
        return self._distinct(self.predicates, *av, **kw)

    def distinct_objects(self, *av, **kw):
        """
        Return a *distinct* set of objects. Arguments are 
        as for :meth:objects
        """
        return self._distinct(self.objects, *av, **kw)

class Graph(_Graph, _Common):
    """
    A :class:`Graph` is a collection of *rdf:Statements*.

    This is the basic :class:`rdflib.graph.Graph` with a few added capabilities:

        * TODO: type and type implementation handling
        * bnode closures
        * handy helper methods :meth:`one` and :meth:`exists` and :meth:`distinct_*`
        * pointer to any :mod:`ordf.handler` in use.

    .. attribute:: identifier
    
    This is the URI of the resource that the graph can be said to be about.

    .. autoattribute:: __types__
    .. autoattribute:: __rules__

    .. attribute:: handler

        If the instance has been obtained via :mod:`ordf.handler`
        it will have the :attr:handler set to the handler that obtained it.

    .. automethod:: bnc
    .. automethod:: one
    .. automethod:: exists
    .. automethod:: replace
    .. automethod:: distinct_subjects
    .. automethod:: distinct_predicates
    .. automethod:: distinct_objects
    """
    
    __types__ = [RDFS["Resource"], RDFG["Graph"]]
    """
    Sub-classes should set this attribute to a list of :class:`URIRef`. It isn't
    necessary to copy data from parent classes. The basic :class:Graph has:

    * *rdfs:Resource*
    * *rdfg:Graph*
    """
    __rules__ = [
        "{ ?s ?p ?o } => { ?s a rdfs:Resource }"
        ]
    """
    Sub-classes should set this attribute to a list of inference relationships
    that are implicit in the nature of their *rdf:type*

    * everything is an *rdfs:Resource*
    """


    def __init__(self, store="IOMemory", identifier=None, **kw):
        if identifier is not None:
            if not isinstance(identifier, Node):
                identifier = URIRef(identifier)
        else:
            identifier = BNode()
        if __rdflib_version__ == 2:
            super(Graph, self).__init__(store, identifier)
        else:
            super(Graph, self).__init__(store=store, identifier=identifier)
        bind_ns(self)

    def version(self):
        version = None
        for s,p,version in self.triples((self.identifier, ORDF.changeSet, None)):
            break
        return version

class ConjunctiveGraph(_CGraph, _Common):
    __types__ = []
    def __init__(self, store="IOMemory", identifier=None, **kw):
        if __rdflib_version__ == 2:
            super(ConjunctiveGraph, self).__init__(store, identifier)
        else:
            super(ConjunctiveGraph, self).__init__(store=store, identifier=identifier, **kw)
        bind_ns(self)

class ReadOnlyGraphAggregate(_ROGraph, _Common):
    pass



"""
.. autoclass:: Agent
"""

__all__ = ["Agent", "Person"]

from ordf.term import BNode, Literal, URIRef
from ordf.graph import Graph
from ordf.namespace import RDF, FOAF
from logging import getLogger
from hashlib import sha1

log = getLogger(__name__)

def rdf_data():
    log.info("Fetching %s" % FOAF)
    g = Graph(identifier=FOAF[""]).parse(FOAF)
    log.info("Parsed %s (%d triples)" % (FOAF, len(g)))
    yield g

def inference_rules(handler, network):
    from FuXi.DLP.DLNormalization import NormalFormReduction
    foaf = handler.get(FOAF[""])
    if len(foaf) == 0:
        for foaf in rdf_data():
            handler.put(foaf)
    NormalFormReduction(foaf)
    return network.setupDescriptionLogicProgramming(foaf, addPDSemantics=False)    

class Agent(Graph):
    """
    .. autoattribute:: __types__
    .. autoattribute:: __rules__

    .. automethod:: nick
    .. automethod:: name
    .. automethod:: homepage
    """
    __types__ = [FOAF["Agent"]]
    "* *foaf:Agent*"
    __rules__ = [
        "{ ?a foaf:homepage ?u . ?b foaf:homepage ?u } => { ?a owl:sameAs ?b }"
        ]
    """
    * Inverse functional property *foaf:homepage* means if two things have 
      the same homepage then they are the same
    """
    def __init__(self, *av, **kw):
        super(Agent, self).__init__(*av, **kw)
        self.add((self.identifier, RDF["type"], FOAF["Agent"]))
    def nick(self, nick):
        """
        Set the *foaf:nick* on the present :class:`Graph`
        """
        self.add((self.identifier, FOAF["nick"], Literal(nick)))
    def name(self, name):
        """
        Set the *foaf:name* on the present :class:`Graph`
        """
        self.add((self.identifier, FOAF["name"], Literal(name)))
    def homepage(self, homepage):
        """
        Set the *foaf:homepage* on the present :class:`Graph`
        """
        self.add((self.identifier, FOAF["homepage"], URIRef(homepage)))
    def mbox(self, email):
        hash = sha1(email)
        self.add((self.identifier, FOAF["mbox_sha1sum"], Literal(hash.hexdigest())))

class Person(Agent):
    __types__ = [FOAF["Person"]]



"""
.. autoclass:: Fresnel
   :show-inheritance:
"""

__all__ = ["Fresnel", "Lens"]

try:
    from xml.etree import ElementTree as ET
except:
    from elementtree import ElementTree as ET
from urllib import urlencode
from ordf.collection import Collection
from ordf.graph import Graph
from ordf.namespace import *
from ordf.term import Literal, URIRef, BNode
from datetime import datetime

log = __import__("logging").getLogger(__name__)

class Lens(Graph):
    """
    """
    def __init__(self, fresnel, identifier):
        super(Lens, self).__init__(identifier=identifier)
        self.fresnel = fresnel
        self.group = None
        self.stylesheets = None

    def properties(self, graph, resource):
        properties = []
        for properties in self.objects(self.identifier, FRESNEL["showProperties"]):
            properties = Collection(self, properties)
            properties = list(properties)
            break
        hideProperties = []
        for hideProperties in self.objects(self.identifier, FRESNEL["hideProperties"]):
            hideProperties = Collection(self, hideProperties)
            hideProperties = list(hideProperties)
            break
        subProperties = []
        for prop in properties:
            if isinstance(prop, BNode) and self.one((prop, RDF["type"], FRESNEL["PropertyDescription"])):
                x = self.one((prop, FRESNEL["sublens"], None))
                if x:
                    lens = self.fresnel.lenses.get(x[2])
                else:
                    lens = None
                for prop in self.objects(prop, FRESNEL["property"]):
                    if prop not in hideProperties:
                        yield prop, lens, self.group.formats.get(prop, Format())
                    subProperties.append(prop)
            elif prop == FRESNEL["allProperties"]:
                for prop in graph.distinct_predicates(resource, None):
                    if prop not in hideProperties and prop not in subProperties and prop not in properties:
                        yield prop, None, self.group.formats.get(prop, Format())
            else:
                if prop not in hideProperties:
                    yield prop, None, self.group.formats.get(prop, Format())

    def formatResource(self, graph, resource):
        if self.group:
            stylesheets = set(self.group.stylesheets)
        else:
            stylesheets = set

        elem = ET.Element("div")
        css = ["fresnel_resource"]

        if isinstance(resource, URIRef):
            elem.set("about", resource)
        elif isinstance(resource, BNode):
            elem.set("about", "[_:%s]" % resource)

        log.debug("    Resource: %s" % resource)
        if isinstance(resource, URIRef):
            uri = ET.Element("div")
            uri.set("class", "fresnel_resource_uri")
            a = ET.Element("a")
            a.set("href", resource)

            a.text = resource
            try:
                qname = self.namespace_manager.qname(resource)
                if not qname.startswith("_"):
                    a.text = qname
            except:
                pass
            uri.append(a)
            elem.append(uri)

        properties = self.properties(graph, resource)
        for property, lens, format in properties:
            styles, prop = self.formatProperty(graph, lens, format, resource, property)
            if prop:
                elem.append(prop)
                [stylesheets.add(x) for x in styles]
        elem.set("class", " ".join(css))
        return stylesheets, elem

    def formatProperty(self, graph, lens, format, resource, property):
        log.debug("      Property: %s" % property)
        log.debug("      Format: %s" % format)

        elem = ET.Element("div")

        css, styles = format.propertyStyles
        css = ["fresnel_property"] + css
        elem.set("class", " ".join(css))
        if styles: elem.set("style", " ".join(styles))

        label = format.label(property)
        if label is not None:
            span = ET.Element("span")
            span.set("class", "fresnel_label")
            a = ET.Element("a")
            a.set("href", property)
            a.text = label
            span.append(a)
            elem.append(span)

        values = [
            self.formatValue(graph, lens, format, resource, property, v)
            for v in graph.objects(resource, property)
        ]

        stylesheets = set()
        if values:
            container = ET.Element("div")
            container.set("class", "fresnel_container")
            [([stylesheets.add(s) for s in sheets], container.append(value)) 
             for (sheets, value) in values]
            elem.append(container)
            elem.set("class", " ".join(css))
        else:
            elem = None
        return stylesheets, elem

    def rst(self, span, value):
        try:
            from docutils import core
            parts = core.publish_parts(source=value, writer_name="html")
            doc = u'<?xml version="1.0" encoding="utf-8"?>\n' + \
                u'<html>\n' + parts["body"] + u"</html>\n"
            elem = ET.XML(doc.encode("utf-8"))
            for c in elem.getchildren():
                span.append(c)
        except ImportError:
            span.text = value

    def formatValue(self, graph, lens, format, resource, property, value):
        elem = ET.Element("div")
        css, styles = format.valueStyles
        css = ["fresnel_value"] + css
        elem.set("class", " ".join(css))
        if styles: elem.set("style", " ".join(styles))

        if format.group:
            stylesheets = set(format.group.stylesheets)
        else:
            stylesheets = set()

        ns, uri, local = self.namespace_manager.compute_qname(property)
        if ns.startswith("_"):
            ns = "ns%s" % ns[1:]
        elem.set("xmlns:%s" % ns, uri)
        pcurie = "%s:%s" % (ns, local)

        if isinstance(value, Literal):
            elem.set("property", pcurie)
            elem.set("content", value)
            if value.language:
                sp = ET.Element("span")
                sp.set("class", "fresnel_object_language")
                sp.text = u" (%s) " % value.language
                elem.append(sp)
                self.rst(elem, value)
                elem.set("xml:lang", value.language)
            elif value.datatype:
                elem.set("xml:lang", "")
                if value.datatype == XSD["dateTime"]:
                    try:
                        dt = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")
                        elem.text = dt.ctime()
                    except:
                        elem.text = value
                else:
                    elem.text = value
                elem.set("datatype", value.datatype)
            else:
                elem.set("xml:lang", "")
                self.rst(elem, value)
        elif isinstance(value, URIRef) or isinstance(value, BNode):
            elem.set("rel", pcurie)
            if isinstance(value, URIRef):
                elem.set("resource", value)
            else:
                elem.set("resource", "[_:%s]" % value)
            if lens:
                styles, rsrc = lens.formatResource(graph, value)
                [stylesheets.add(x) for x in styles]
                elem.append(rsrc)
            elif isinstance(value, BNode):
                lens = DefaultLens(self.fresnel, self.group)
                styles, rsrc = lens.formatResource(graph, value)
                [stylesheets.add(x) for x in styles]
                elem.append(rsrc)
            else:
                if isinstance(value, URIRef):
                    if format.one((format.identifier, FRESNEL["value"], FRESNEL["image"])):
                        a = ET.Element("img")
                        a.set("src", value)
                        try:
                            qname = self.namespace_manager.qname(property)
                            a.set("alt", "qname")
                        except:
                            a.set("alt", property)
                    else:
#                        if format.one((format.identifier, FRESNEL["value"], FRESNEL["externalLink"])):
                        a = ET.Element("a")
                        a.set("href", value)
                        try:
                            pfx, uri, rest = self.compute_qname(value)
                            if not pfx.startswith("_"):
                                value = u"%s:%s" % (pfx, rest)
                        except:
                            pass
                        a.text = value
                else:
                    a = ET.Element("span")
                    a.text = value
                elem.append(a)
        return stylesheets, elem

class DefaultLens(Lens):
    def __init__(self, fresnel, group):
        super(DefaultLens, self).__init__(fresnel, None)
        self.group = group
    def properties(self, graph, resource):
        for prop in graph.distinct_predicates(resource):
            yield prop, None, self.group.formats.get(prop, Format())

class Group(Graph):
    def __init__(self, *av, **kw):
        super(Group, self).__init__(*av, **kw)
        self.lenses = {}
        self.formats = {}
        
    @property
    def stylesheets(self):
        if not hasattr(self, "__stylesheets__"):
            self.__stylesheets__ = list(self.distinct_objects(self.identifier, FRESNEL["stylesheetLink"]))
        return self.__stylesheets__

class Format(Graph):
    def __init__(self, *av, **kw):
        super(Format, self).__init__(*av, **kw)
        self.group = None
    def label(self, property):
        # what spaghetti is this!
        for _s, _p, label in self.triples((self.identifier, FRESNEL["label"], None)):
            if label == FRESNEL["none"]:
                return None
            else:
                return label
        try:
            qname = self.namespace_manager.qname(property)
            if not qname.startswith("_"):
                return qname
        except:
            pass
        label = ":" + property.rsplit("/", 1)[-1].rsplit("#", 1)[-1]
        return label

    def styles(self, predicate):
        css, styles = [], []
        for s in self.objects(self.identifier, predicate):
            if s.datatype == FRESNEL["stylingInstructions"]:
                styles.append(s)
            elif s.datatype == FRESNEL["styleClass"]:
                css.append(s)
        return css, styles

    @property
    def propertyStyles(self):
        return self.styles(FRESNEL["propertyStyle"])
    @property
    def valueStyles(self):
        return self.styles(FRESNEL["valueStyle"])

class Fresnel(Graph):
    """
    .. automethod:: format
    """
    __types__ = []
    def __init__(self, *av, **kw):
        super(Fresnel, self).__init__(*av, **kw)
        self.lenses = {}
        self.classLenses = {}
        self.instanceLenses = {}
        self.groups = {}
        self._compiled = False

    def compile(self):
        if self._compiled:
            return
        self._compiled = True

        for l in self.distinct_subjects(RDF["type"], FRESNEL["Lens"]):
            lens = Lens(fresnel=self, identifier=l) 
            lens += self.bnc((l, None, None))
            self.lenses[l] = lens
            for cls in lens.objects(l, FRESNEL["classLensDomain"]):
                self.classLenses[cls] = lens
            for cls in lens.objects(l, FRESNEL["instanceLensDomain"]):
                self.instanceLenses[cls] = lens
            for grp in lens.objects(l, FRESNEL["group"]):
                group = self.groups.get(grp)
                if group is None:
                    group = Group(identifier=grp)
                    group += self.bnc((grp, None, None))
                    self.groups[grp] = group
                group.lenses[l] = lens
                lens.group = group

        for fmt in self.distinct_subjects(RDF["type"], FRESNEL["Format"]):
            format = Format(identifier=fmt)
            format += self.bnc((fmt, None, None))
            for grp in format.objects(fmt, FRESNEL["group"]):
                group = self.groups.get(grp)
                if group is None:
                    group = Group(identifier=grp)
                    group += self.bnc((grp, None, None))
                    self.groups[grp] = group
                for dom in format.objects(fmt, FRESNEL["propertyFormatDomain"]):
                    group.formats[dom] = format
                format.group = group

    def format(self, graph, resource=None):
        """
        Format the given graph according to this set of fresnel instructions.

        :return: tuple (stylesheet list, elementtree document)
        """
        self.compile()
        stylesheets = set()
        root = ET.Element("div")
        root.set("class", "fresnel_container")

        _x = self.one((self.identifier, ORDF["lensList"], None))
        if _x: lenses = Collection(self, _x[2])
        else: lenses = self.lenses

        for lens_uri in lenses:
            log.debug("Lens: %s" % lens_uri)
            lens = self.lenses.get(lens_uri)
            if lens is None:
                continue
            for typ in lens.distinct_objects(lens.identifier, FRESNEL["classLensDomain"]):
                log.debug("    Type: %s" % typ)
                resources = list(graph.triples((resource, RDF["type"], typ)))
                resources.sort(lambda x,y: cmp(x[0], y[0]))
                for resource, _p, _o in resources:
                    styles, element = lens.formatResource(graph, resource)
                    [stylesheets.add(x) for x in styles]
                    root.append(element)

        return stylesheets, root

from ordf.command import Command
class Render(Command):
    parser = Command.StandardParser(usage="%prog [options]")
    parser.add_option("-t", "--template", 
                      dest="template",
                      help="HTML Template")
    parser.add_option("-r", "--resource",
                      dest="resource",
                      help="Resource to format")
    default_template = """\
<html>
<body>
<fresnel />
</body>
</html>"""
    def command(self):
        template = self.default_template
        if self.options.template:
            template = open(self.options.template).read()
        template = ET.fromstring(template)

        fresnel_uri, graph_uri = self.args
        tmp = self.handler.get(fresnel_uri)
        fresnel = Fresnel(store=tmp.store, identifier=tmp.identifier)
        graph = self.handler.get(graph_uri)
        stylesheets, tree = fresnel.format(graph, URIRef(self.options.resource))

        head = template.find("head")
        if head is None:
            head = ET.Element("head")
            template.insert(0, head)
        for sheet in stylesheets:
            link = ET.Element("link")
            link.set("rel", "stylesheet")
            link.set("href", sheet)
            link.set("type", "text/css")
            head.append(link)
        meta = ET.Element("meta")
        meta.set("http-equiv", "Content-Type")
        meta.set("content", "text/html; charset=utf-8")
        head.append(meta)

        # guess at a good title for the document
        resource = URIRef(self.options.resource) if self.options.resource else graph.identifier
        title = None
        for label in DC["title"], FOAF["name"], RDFS["label"]:
            for title in graph.distinct_objects(resource, label):
                break
            if title is not None:
                break
        if title is None:
            title = resource
        tt = ET.Element("title")
        tt.text = title
        head.append(tt)

        def find_fresnel(parent):
            index = 0
            for child in parent:
#                if child.tag == "{http://www.w3.org/1999/xhtml}fresnel":
                if child.tag == "fresnel":
                    return parent, child, index
                next = find_fresnel(child)
                if next is not None:
                    return next
                index += 1
        parent, replace, index = find_fresnel(template)
        parent.remove(replace)
        parent.insert(index, tree)

        doc = '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML+RDFa 1.0//EN" "http://www.w3.org/MarkUp/DTD/xhtml-rdfa-1.dtd">'
        doc += '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">'
        for child in template:
            doc += ET.tostring(child, encoding="utf-8")
        doc += '</html>'

        print doc

    #    from xml.dom import minidom
    #    from StringIO import StringIO
    #    doc = minidom.parse(StringIO(doc))
    #    print doc.toprettyxml()

def render():
    Render().command()



from ordf.vocab.fresnel import Fresnel
from ordf.tests import fixture
from sys import stdout
from xml.dom import minidom

def lens():
    g = fixture("lens.n3")
    return Fresnel(store=g.store, identifier=g.identifier)

def test_01_lens():
    l = lens()
    g = fixture("marc_bio.n3")
    styles, doc = l.format(g)
    print minidom.parseString(doc).toprettyxml()



from ordf.vocab.owl import predicate, AnnotatibleTerms, Class
from ordf.namespace import DC, FOAF, OWL, RDF, RDFS, VOID

class Dataset(Class):
    def __init__(self, *av, **kw):
        super(Dataset, self).__init__(VOID.Dataset, *av, **kw)
    def get(self, identifier, graph=None):
        if graph is None: graph=self.graph
        ds = _Dataset(identifier, graph=graph)
        ds.factoryGraph = ds.graph
        ds.type = self.identifier
        return ds
    
class _Dataset(AnnotatibleTerms):
    # xxx should we put DC and FOAF things like this somewhere
    # in a superclass? hard to do without inappropriately
    # hardwiring a class hierarchy
    title = predicate(DC["title"])
    description = predicate(DC.description)
    isPartOf = predicate(DC.isPartOf)
    homepage = predicate(FOAF.homepage)
    page = predicate(FOAF.page)

    subset = predicate(VOID.subset)
    dataDump = predicate(VOID.dataDump)
    vocabulary = predicate(VOID.vocabulary)
    sparqlEndpoint = predicate(VOID.sparqlEndpoint)
    exampleResource = predicate(VOID.exampleResource)



__all__ = ["RDFObj"]

from ordf.term import URIRef
from ordf.namespace import RDF
from pprint import pformat

class RDFObj(object):
    def __init__(self, graph, subject):
        self.graph = graph
        self.subject = subject

    def __str__(self):
        types = [str(o) for s,p,o in self.graph.triples((self.subject, RDF.type, None))]
        dct = {
            "graph": self.graph.identifier,
            "subject": self.subject,
            "types": types
        }
        return pformat(dct)
    def __repr__(self):
        return str(self)

    def dict(self):
        dct = {}
        for s,p,o in self.graph.triples((self.subject, None, None)):
            dct.setdefault(p, []).append(o)
        return dct

    def keys(self):
        return self.dict().keys()
    def values(self):
        return self.dict().values()
    def items(self):
        return self.dict().items()

    def __getitem__(self, k):
        def iterpred():
            for s,p,o in self.graph.triples((self.subject, k, None)):
                if isinstance(o, URIRef):
                    yield RDFObj(self.graph, o)
                else:
                    yield o
        return list(iterpred())



"""
Changeset Tests
"""

from ordf.graph import Graph
from ordf.term import URIRef
from ordf.vocab.changeset import ChangeSet
from ordf.namespace import namespaces, RDF, CS
try:
    from rdflib.store.IOMemory import IOMemory
except ImportError:
    from rdflib.plugins.memory import IOMemory
from ordf.tests import our_test_graph, card_graph, identifier, nosep
import os

store = IOMemory()

def test_00_null():
    orig = our_test_graph()
    new = our_test_graph()

    ## make a changeset from nothing to our test data, and save
    cs = ChangeSet("test_cs", "Testing Fixtures")
    assert cs.diff(orig, new) == 0
    cs.commit()
    assert not cs

def test_01_create():
    orig = Graph(identifier=identifier)
    new = our_test_graph()

    ## make a changeset from nothing to our test data, and save
    cs = ChangeSet("test_cs", "Testing Fixtures")
    assert cs.diff(orig, new) == 85
    cs.commit()
    assert cs

def test_02_remove():
    ## create from empty
    orig = Graph(identifier=identifier)
    new = our_test_graph()
    cs = ChangeSet("test_cs", "Creating Graph", store=store)
    assert cs.diff(orig, new) == 85
    cs.commit()
    assert cs

    ## delete the data and check
    removed = our_test_graph()
    removed.remove((None, RDF.type, None))
    cs = ChangeSet("test_cs", "Removing Types", store=store)
    assert cs.diff(new, removed) == 18
    cs.commit()
    assert cs

    ## put the data back
    cs = ChangeSet("test_cs", "Adding Types Back", store=store)
    assert cs.diff(removed, our_test_graph()) == 18
    cs.commit()
    assert cs

def test_03_three():
    ## operations on three graphs in one changeset
    g1 = our_test_graph()
    g2 = card_graph()
    g3 = Graph(identifier="http://example.org/empty")
    g4 = Graph(identifier="http://example.org/empty")

    cs = ChangeSet("test_cs", "Delete test add card")
    assert cs.diff(g1, Graph(identifier=g1.identifier)) == 85
    assert cs.diff(Graph(identifier=g2.identifier), g2) == 61
    assert cs.diff(g3, g4) == 0
    cs.commit()

    changed = []
    for s,p,o in cs.triples((cs.identifier, CS.subjectOfChange, None)):
        changed.append(o)
    assert len(changed) == 2



import os
import tempfile
import shutil

from ordf.handler import Handler, HandlerPlugin
from ordf.namespace import RDF, Namespace
from ordf.graph import Graph
from ordf.tests import our_test_graph, identifier

class TestHandler:
    @classmethod
    def setup_class(self):
        self.tmpdir = os.path.join(tempfile.gettempdir(), self.__name__)
        ptree_cls = HandlerPlugin.find("pairtree")
        ptree = ptree_cls(self.tmpdir)
        self.handler = Handler(namespace=Namespace("http://example.org/changesets/"))
        self.handler.register_reader(ptree)
        self.handler.register_writer(ptree)

    @classmethod
    def teardown_class(self):
        shutil.rmtree(self.tmpdir)

    def test_01_find(self):
        assert HandlerPlugin.find("null") is not None

    def test_02_register(self):
        h = Handler()

        # test  direct registering of an instance
        r = HandlerPlugin()
        h.register_reader(r)

        # test registering by named plugin
        h.register_writer("null")


    def test_03_context(self):
        ctx = self.handler.context("test_handler", "create some data")
        ctx.add(our_test_graph())
        ctx.commit()

        g = our_test_graph()
        g.remove((None, RDF.type, None))
        ctx = self.handler.context("test_handler", "delete some data")
        ctx.add(g)
        ctx.commit()

    def test_04_undo(self):
        ## undo the last change (remove the types again)
        constructed = self.handler.get(identifier)
        ver = constructed.version()
        cs = self.handler.changeset(ver)
        cs.undo(constructed)

        ctx = self.handler.context("test_handler", "put data back")
        ctx.add(our_test_graph())
        ctx.commit()

    def test_05_construct(self):
        ## construct head from changes in previous step
        constructed = self.handler.construct(identifier)
        assert len(constructed) == len(our_test_graph()) + 1




"""
RDF Proxy Service
-----------------

This controller is mainly intended for use by JavaScript applications that
are not able to connect to remote hosts and so cannot download RDF data from
them directly. As with L{ordf.pylons.graph} it takes the parameters
B{uri} and B{format} specifying the (remote) URI to fetch and the format
in which to return the results.

The controller will request the remote URI with I{application/rdf+xml} which
is the only official RDF serialisation, will parse it using L{ordf.graph.Graph}
and then serialise the results as requested.
"""
import logging
from cStringIO import StringIO

from pylons import request, response, session, tmpl_context as c
from pylons.controllers.util import abort
from pylons import config

from ordf.graph import Graph
from ordf.term import URIRef
from ordf.pylons import ControllerFactory
from ordf.pylons.serialiser import graph_format

log = logging.getLogger(__name__)

def ProxyControllerFactory(base, model):
    return ControllerFactory("ProxyController", _ProxyController, base, model)

class _ProxyController:
    def index(self):
        accept, format = graph_format(request)

        uri = request.GET.get("uri", None)
        if uri is None:
            c.message = "Missing GET parameter: uri"
            return self.render("proxy.html")

        graph = Graph(identifier=URIRef(uri))
        try:
            graph.parse(uri)
        except:
            response.content_type = "text/plain"
            return "Unable to parse or fetch %s" % (uri,)

        if format == "html":
            c.triples = graph.triples((None, None, None))
            data = self.render("graph.mako")
        else:
            data = graph.serialize(format=format)
            response.content_type = str(accept)
        return data



#from ordf.handler import HandlerPlugin
#from ordf.namespace import RDFS
#import redis
#
#class Redis(HandlerPlugin):
#    def __init__(self, *av, **kw):
#        self.db = redis.Redis(*av, **kw)
#    def cset(self, cs):
#        self.db.lpush("RecentChanges", cs.metadata.serialize(format="n3"))
#        if self.db.llen("RecentChanges") > 10:
#            self.db.rpop("RecentChanges")



"""
.. program:: ordf

:program:`ordf`
===============

A command line program, :program:`ordf`, for running handlers and processing 
messages. This is somewhat of a swiss-army knife but some of the more
common use cases are described below.

.. cmdoption:: -c config.ini, --config config.ini

    Some command line options can be specified in the configuration file
    with a similar syntax. Other than for testing it is almost always going
    to be desirable to run with a configuration file.
    
    Any options given on the command line will override the corresponding
    configuration file parameters.

    The configuration file should have a section *[app:main]* for 
    compatibility with pylons configuration files. This configuration 
    section is passed to :func:`ordf.handler.init_handler`

.. cmdoption:: -r reader[,reader[...]], --readers reader[,reader[...]]

    A comma-separated list of reader plugins. Equivalent to *ordf.readers*
    in the configuration file.

.. cmdoption:: -w writer[,writer[...]], --writers writer[,writer[...]]

    A comma-separated list of writer plugins. Equivalent to *ordf.writers*
    in the configuration file.

.. cmdoption:: -l logfile, --logfile logfile

    A file to write log messages to.

.. cmdoption:: -v verbosity, --verbosity verbosity

    The minimum level of log messages to write to the file. Must be one of
    *debug, info, warning, error* or *critical*.

.. cmdoption:: -s, --save

    Save the arguments to the store.

.. cmdoption:: -x, --remove

    Remove the arguments from the store by replacing them with an empty graph.

.. cmdoption:: -f format, --format format

    For operations that involve reading RDF data, the format that it is in.
    This must be one of the formats supported by rdflib. The default is XML.

.. cmdoption:: -i uri, --identifier uri

    For operations that involve writing data to the store, the graph identifier 
    that should be the destination of the data. Default is the filename.

.. cmdoption:: -u username, --user username

    Username for logging changes

.. cmdoption:: -m message, --message message

    Log message or change reason for logging changes

.. cmdoption:: -d, --daemon

    Run as a daemon, listening to a source and writing data to local storage.

.. cmdoption:: --reload

    Run in a reloading file monitor, reloading if changes to python code or
    configuration file are made.

.. cmdoption:: --reindex

    Rebuild indices iterating over a single reader

.. program:: ordf_load

:program:`ordf_load`
=======================

Utility program for loading any RDF fixtures that are packaged with some
python modules. Normal usage is as in::

    % ordf_load -p ordf.vocab
    INFO  [ordf.handler] Handler(0/0) initialised ver 0.7.246.1a3933607282
    INFO  [ordf.handler] Handler(1/0) reading from <ordf.handler.pt.PairTree object ...>
    INFO  [ordf.handler] Handler(1/1) writing to <ordf.handler.pt.PairTree object ...>
    INFO  [ordf.handler.rdf.RDFLib] Initialising Sleepycat storage
    INFO  [ordf.handler] Handler(1/2) writing to <ordf.handler.rdf.RDFLib object ...>
    INFO  [ordf.load/3149] Loading http://ordf.org/schema/ordf
    INFO  [ordf.changeset] 20 changes urn:uuid:44764ed0-8d12-11df-acde-001f5bef60ee

    % ordf_load -p ordf.onto
    INFO  [ordf.handler] Handler(0/0) initialised ver 0.7.246.1a3933607282
    INFO  [ordf.handler] Handler(1/0) reading from <ordf.handler.pt.PairTree ...>
    INFO  [ordf.handler] Handler(1/1) writing to <ordf.handler.pt.PairTree ...>
    INFO  [ordf.handler.rdf.RDFLib] Initialising Sleepycat storage
    INFO  [ordf.handler] Handler(1/2) writing to <ordf.handler.rdf.RDFLib ...>
    INFO  [ordf.load/3148] Loading http://ordf.org/lens/changeset
    INFO  [ordf.load/3148] Loading http://ordf.org/lens/fresnel
    INFO  [ordf.load/3148] Loading http://ordf.org/lens/ontology
    INFO  [ordf.load/3148] Loading http://ordf.org/lens/rdfs
    INFO  [ordf.changeset] 612 changes urn:uuid:171ffd8c-8d12-11df-850a-001f5bef60ee

In addition to the *-c* for specifying the config file, *-l* and *-v*
for log file and debugging, and *-r* and *-w* for specifying readers and
writers as with :program:`ordf`, :program:`ordf_load` supports these
arguments:

.. cmdoption:: -p, --package

    Name of the package (python module) to look for N3 fixtures in. This
    doesn't work with namespace packages (e.g. ordf itself which is why
    the schema fixtures are in the ordf.vocab sub-module).

Command Implementation -- :class:`Command`
==========================================

.. autoclass:: Command
"""
from ordf.graph import Graph
from ordf.handler import init_handler
from ordf.namespace import RDF, RDFS
from ordf.term import Literal, URIRef
from ordf.vocab.opmv import Agent, Process
from optparse import OptionParser
from ConfigParser import ConfigParser
from getpass import getuser
from paste import reloader
from glob import glob
import subprocess, signal
import pkg_resources
import os, sys
import urllib
import logging
import sys

def config(filename):
    cfgpath = os.path.abspath(filename)
    cfgfile = ConfigParser({ "here": os.path.dirname(cfgpath) })
    cfgfile.read(cfgpath)

    cfg = {}
    if cfgfile.has_section("app:main"):
        cfg.update(cfgfile.items("app:main"))

    return cfg

class Command(object):
    """
    This class is very similar to :class:`paste.script.command.Command` but
    rather than implementing a :program:`paster` plugin it is for stand-alone
    command line programs. To implement a command line program, sub-class this
    class, and make a minimal method to instantiate and run it. As with the
    paster counterpart you have to add an option parser and a method called 
    :meth:`command`. A minimal example:

    .. code-block:: python

        class Hello(Command):
            parser = Command.StandardParser(usage="%prog [options]")
            def command(self):
                print "hello world"

        def hello():
            Hello().command()

    To create the actual script, in your package's *setup.py* needs an entry
    point like::

        [console_scripts]
        hello=mypackage.command:hello

    and then run one of::

        % python setup.py develop
        % python setup.py install
    """
    def __init__(self):
        self.parse_args()
        self.parse_config()
        self.setup_logging()
        self.setup_handler()

    @classmethod
    def StandardParser(cls, *av, **kw):
        parser = OptionParser(*av, **kw)
        parser.add_option("-c", "--config",
                          dest="config", default="development.ini",
                          help="configuration file (default: development.ini)")
        parser.add_option("-l", "--logfile",
                          dest="logfile", default=None,
                          help="log to file")
        parser.add_option("-v", "--verbosity",
                          dest="verbosity", default="info",
                          help="log verbosity. one of debug, info, warning, error, critical")
        return parser

    def parse_args(self):
        self.options, self.args = self.parser.parse_args()

    def parse_config(self):
        self.config = {}

        if self.options.config:
            cfg = config(self.options.config)
            self.config.update(cfg)

    def setup_logging(self):
        ## set up logging
        logcfg = { 
            "level": logging.INFO,
            "format": "%(levelname)s  [%(name)s] %(message)s"
            }
        if self.options.logfile:
            logcfg["filename"] = self.options.logfile
            logcfg["format"] = "%(asctime)s %(levelname)s  [%(name)s] %(message)s"
        if self.options.verbosity:
            levels = {
                "debug": logging.DEBUG,
                "info": logging.INFO,
                "warning": logging.WARNING,
                "error": logging.ERROR,
                "critical": logging.CRITICAL
                }
            logcfg["level"] = levels.get(self.options.verbosity, logging.NOTSET)
        logging.basicConfig(**logcfg)

    def setup_handler(self):
        self.handler = init_handler(self.config)

class ORDF(Command):
    usage = "Usage: %prog [options] [files...]"
    _reloader_environ_key = 'PYTHON_RELOADER_SHOULD_RUN'
    parser = Command.StandardParser(usage=usage)
    parser.add_option("-r", "--readers",
                       dest="readers", default=None,
                       help="comma separated list of reader plugins")
    parser.add_option("-w", "--writers",
                       dest="writers", default=None,
                       help="comma separated list of writer plugins")
    parser.add_option("-s", "--save", action="store_true",
                       dest="save", default=False,
                       help="save the arguments to the store")
    parser.add_option("-f", "--format",
                       dest="format", default="xml",
                       help="format of rdf files for read operations. default XML")
    parser.add_option("-i", "--identifier",
                       dest="identifier", default=None,
                       help="the graph uri to store the data")
    parser.add_option("-u", "--user",
                       dest="user", default=None,
                       help="username for logging. defaults to environment")
    parser.add_option("-m", "--message",
                       dest="message", default=None,
                       help="change message for logging")
    parser.add_option("-x", "--remove", action="store_true",
                        dest="remove", default=False,
                        help="remove graph from store")
    parser.add_option("-d", "--daemon", action="store_true",
                       dest="daemon", default=False,
                       help="Do not exit, loop and process messages")
    parser.add_option("--reload", action="store_true",
                       dest="reload", default=False,
                       help="Check files for changes and reload")
    parser.add_option("--reindex", action="store_true",
                       dest="reindex", default=False,
                       help="Reindex indices reading from a single reader")
    parser.add_option("--nocs", action="store_true",
                      dest="nocs", default=False,
                      help="Do not process changesets")
    def parse_config(self):
        super(ORDF, self).parse_config()

        ## override config with the command line
        if self.options.readers:
            self.config["ordf.readers"] = self.options.readers
        if self.options.writers:
            self.config["ordf.writers"] = self.options.writers

        if not self.options.nocs:
            if self.options.save or self.options.remove:
                if not self.options.message:
                    self.parser.print_help()
                    print "ERROR: missing log message"
                    exit(1)

        if self.options.reindex:
            reader = self.config.get("ordf.readers")
            if reader.find(",") >= 0:
                print "--reindex requires exactly one reader (-r/--readers)"
                exit(1)
            self.reindex_reader = reader

    def setup_logging(self):
        super(ORDF, self).setup_logging()
        self.log = logging.getLogger("ordf.command/%s" % os.getpid())

    def command(self):
        if self.options.reload:
            if os.environ.get(self._reloader_environ_key):
                self.log.debug("using reloading file monitor")
                reloader.install(1)
                if self.options.config:
                    reloader.watch_file(self.options.config)
            else:
                return self.restart_with_reloader()

        if self.options.save or self.options.remove:
            if self.options.user:
                user = self.options.user
            else:
                user = getuser()

            if not self.options.nocs:
                ctx = self.handler.context(user, self.options.message)

            agent = Agent()
            agent.nick(Literal(user))

            for filename in self.args:

                if self.options.identifier:
                    ident = URIRef(self.options.identifier)
                else:
                    proto, rest = urllib.splittype(filename)
                    if proto:
                        ident = URIRef(filename)
                    else:
                        ident = URIRef("file://" + os.path.abspath(filename))

                g = Graph(identifier=ident)

                if self.options.save:
                    g.parse(filename, format=self.options.format)

                    proc = Process()
                    proc.agent(agent)
                    proto, rest = urllib.splittype(filename)
        
                    if proto:
                        src = URIRef(filename)
                    else:
                        src = URIRef("file://" + os.path.abspath(filename))
                    proc.use(src)
                    proc.result(g)

                    self.log.info("add %s" % (g.identifier,))
                    if self.options.nocs:
                        self.handler.put(g)

                elif self.options.remove:
                    self.log.info("del %s" % (g.identifier,))
                    if self.options.nocs:
                        self.handler.remove(g)

                if not self.options.nocs:
                    ctx.add(g)

            if not self.options.nocs:
                self.log.info("commit changes")
                cs = ctx.commit()

        elif self.options.reindex:
            reader = getattr(self.handler, self.reindex_reader)
            for graph in reader:
                self.log.info("indexing %s" % graph.identifier)
                self.handler.put(graph)
        else:
            for uri in self.args:
                g = self.handler.get(uri)
                print g.serialize(format=self.options.format)
        
        def _daemon():
            from time import sleep
            try:
                while True:
                    sleep(1)
            except KeyboardInterrupt:
                pass

        if self.options.daemon:
            _daemon()
        self.handler.close()

    def restart_with_reloader(self):
        while 1:
            args = [sys.executable] + sys.argv
            new_environ = os.environ.copy()
            new_environ[self._reloader_environ_key] = 'true'
            proc = None
            try:
                try:
                    def handle_term(signo, frame):
                        raise SystemExit
                    signal.signal(signal.SIGTERM, handle_term)
                    proc = subprocess.Popen(args, env=new_environ)
                    self.log.debug("started child process %s" % proc.pid)
                    exit_code = proc.wait()
                    proc = None
                except KeyboardInterrupt:
                    self.log.info('^C caught in monitor process')
                    return 1
            finally:
                if (proc is not None
                    and hasattr(os, 'kill')):
                    try:
                        os.kill(proc.pid, signal.SIGTERM)
                    except (OSError, IOError):
                        pass
            if exit_code != 3:
                return exit_code
            self.log.info('-'*20 + ' Restarting ' + '-'*20)

class ORDF_LOAD(Command):
    usage = "%prog [options]"
    parser = Command.StandardParser(usage=usage)
    parser.add_option("-p", "--package",
                      dest="package_name",
                      default="ordf.vocab",
                      help="Package name to read N3 fixtures from (default: ordf.vocab) " +
                           "-- does not work for namespace packages")
    parser.add_option("--nocs",
                      dest="nocs",
                      action="store_true",
                      default=False,
                      help="Put data directly into store without changeset")
    parser.add_option("-r", "--readers",
                       dest="readers", default=None,
                       help="comma separated list of reader plugins")
    parser.add_option("-w", "--writers",
                       dest="writers", default=None,
                       help="comma separated list of writer plugins")

    def parse_config(self):
        super(ORDF_LOAD, self).parse_config()

        ## override config with the command line
        if self.options.readers:
            self.config["ordf.readers"] = self.options.readers
        if self.options.writers:
            self.config["ordf.writers"] = self.options.writers

    def setup_logging(self):
        super(ORDF_LOAD, self).setup_logging()
        self.log = logging.getLogger("ordf.load/%s" % os.getpid())

    def command(self):
        pkg = __import__(self.options.package_name, globals(), locals(), ["version"])

        ctx = self.handler.context(getuser(), "ORDF N3 Import")

        if hasattr(pkg, "rdf_data"):
            for graph in pkg.rdf_data():
                label = graph.one((graph.identifier, RDFS["label"], None))
                if label is not None:
                    self.log.info("Loading %s (%s)" % (graph.identifier, label[2]))
                else:
                    self.log.info("Loading %s" % (graph.identifier,))
                if self.options.nocs:
                    self.handler.put(graph)
                else:
                    ctx.add(graph)

        if not self.options.nocs:
            ctx.commit()

def ordf():
    ORDF().command()

def load_rdf():
    ORDF_LOAD().command()



"""
This module implements some triplestores. Most basic but least 
scalable is the storage that comes with `RDFLib`_. More advanced
and scalable but more complicated to set up is `4store`_.

RDFLib Basic Storage -- :class:`RDFLib`
---------------------------------------
.. autoclass:: RDFLib
   :show-inheritance:

4Store Quad-Store -- :class:`FourStore`
---------------------------------------
.. autoclass:: FourStore
   :show-inheritance:

.. _RDFLib: http://www.rdflib.net/
.. _4store: http://4store.org/
"""

__all__ = ["RDFLib", "FourStore"]

from rdflib import plugin
from rdflib.store import Store
from ordf.graph import Graph, _Graph, ConjunctiveGraph
from ordf.utils import get_identifier
from ordf.handler import HandlerPlugin
from logging import getLogger

class RDFLib(HandlerPlugin):
    """
    The RDFLib handler can use any of the back-ends that are supported
    by rdflib itself. The constructor takes a *store* keyword argument
    that specifies which to use. This defaults to *"IOMemory"* which is
    not terribly useful. Initialisation arguments for the store should
    be passed as positional arguments, for example::

        RDFLib("/some/where/data", store="Sleepycat")

    or the equivalent in the configuration file,::

        rdflib.args = /some/where/data
        rdflib.store = Sleepycat
    """
    def __init__(self, *av, **kw):
        kw = kw.copy()
        store = kw.setdefault("store", "IOMemory")
        del kw["store"]
        self.log = getLogger(__name__ + "." + self.__class__.__name__)
        if isinstance(store, basestring):
            store_cls = plugin.get(store, Store)
            self.log.info("Initialising %s storage" % store_cls.__name__)
            store = store_cls(*av, **kw)
        self.store = store
    def __getitem__(self, k):
        ## Careful, the graph in the store is to be treated 
        ## as read-only unless a set operation is made.
        ## Unfortunately this means we have to copy it
        ident = get_identifier(k)
        orig = Graph(self.store, identifier=ident)
        new = Graph(identifier=ident)
        for statement in orig.triples((None, None, None)):
            new.add(statement)
        return new
    def __setitem__(self, k, g):
        assert isinstance(g, _Graph)
        old = Graph(self.store, identifier=get_identifier(k))
        old.remove((None, None, None))
        for statement in g.triples((None, None, None)):
            old.add(statement)
        self.store.commit()
        if hasattr(self.store, "sync"):
            self.store.sync()
    def __delitem__(self, k):
        old = Graph(self.store, identifier=get_identifier(k))
        old.remove((None, None, None))
        self.store.commit()
        if hasattr(self.store, "sync"):
            self.store.sync()

    def append(self, g):
        old = Graph(self.store, identifier=g.identifier)
        old += g
        if hasattr(self.store, "sync"):
            self.store.sync()
                                    
    def remove(self, g):
        del self[g]

    def query(self, *av, **kw):
        g = ConjunctiveGraph(self.store)
        return g.query(*av, **kw)

class FourStore(RDFLib):
    """
    Use of this back-end requires the `py4s`_ bindings to `4store`_.
    It also requires 4store to be built from the `multiclient branch`_
    to support multiple simultaneous client connections.

    Initialisation takes a comma-separated string of arguments,
    only the first, the name of the *kb* is required. For example::

        FourStore("kbname")
        FourStore("kbname,soft_limit=-1")

    or equivalently in the configuration file::

        fourstore.args = kbname
        fourstore.args = kbname,soft_limit=-1

    .. _py4s: http://github.com/wwaites/py4s
    .. _multiclient branch: http://github.com/wwaites/4store
    """
    def __init__(self, config):
        from py4s import LazyFourStore
        self.store = LazyFourStore(config)
    def __setitem__(self, k, g):
        assert isinstance(g, _Graph)
        ident = get_identifier(k)
        cursor = self.store.cursor()
        cursor.add_graph(g, replace=True)
    def __delitem__(self, k):
        ident = get_identifier(k)
        cursor = self.store.cursor()
        cursor.delete_graph(ident)
    def append(self, frag):
        ident = get_identifier(frag)
        cursor = self.store.cursor()
        cursor.add_graph(frag, replace=False)
    def query(self, *av, **kw):
        return self.store.query(*av, **kw)



from pylons import request, response, session, tmpl_context as c
from ordf.pylons.serialiser import search_format
from ordf.pylons import ControllerFactory
from ordf.graph import Graph
from cStringIO import StringIO
import pkg_resources
try:
    import json
except ImportError:
    import simplejson as json

class _XapianController:
    def index(self):
        if not hasattr(self, "search"):
            for entrypoint in pkg_resources.iter_entry_points(group="ordf.xapian"):
                if entrypoint.name == "search":
                    self.search = entrypoint.load()
                    break
        if not hasattr(self, "search"):
            raise ImportError("no plugin in [ordf.xapian] named %s" % self.name)

        c.query = request.GET.get("q", None)
        if c.query:
            c.offset = int(request.GET.get("offset", 0)) if not hasattr(c, 'offset') else c.offset
            c.limit = int(request.GET.get("limit", 20)) if not hasattr(c, 'limit') else c.limit
            c.results = self.search(self.handler.xapian.open(), c.query, c.offset, c.limit)

        mime_type, format = search_format(request)

        if format == "json":
            resp = []
            for row in c.results:
                g = Graph()
                g.parse(StringIO(row.document.get_data()), format="n3")
                r = {
                    "metadata": {
                        "rank": row.rank,
                        "percent": row.percent,
                        "weight": row.weight,
                    },
                    "uri": row.document.get_value(0),
                    "document": g.serialize(format="nt"),
                }
                resp.append(r)
            resp = json.dumps(resp)
        else:
            resp = self._render()
        if c.query:
            self.handler.xapian.close()
        return resp

    def _render(self):
        return self.render("search.html")

def XapianControllerFactory(base, handler):
    return ControllerFactory("XapianController", _XapianController, base, handler)

